import torch
import numpy as np
import copy
import sys
import time

import torch.distributed as dist

from metroem import loss

from pdb import set_trace as st

def align_sample(model, bundle, train=False):
    tgt = bundle['tgt'].unsqueeze(1)
    src = bundle['src'].unsqueeze(1)

    src_field = None
    if 'src_field' in bundle:
        src_field = bundle['src_field']
    pred_res, state = model.forward(src_img=src, tgt_img=tgt, src_agg_field=src_field,
            return_state=True, train=train)

    bundle['pred_tgt'] = pred_res.from_pixels()(src)
    bundle['pred_res'] = pred_res

    bundle['state'] = state
    #print (pred_res[..., (src != 0).squeeze()].abs().mean())
    return bundle


def aligner_train_loop(rank,
                       model,
                       mip_in,
                       train_loader,
                       val_loader,
                       optimizer,
                       print_every=50,
                       num_epochs=1000,
                       loss_fn=None,
                       reverse=True,
                       manual_run_transform_def={},
                       manual_loss_transform_def={},
                       checkpoint_folder='./checkpoints',
                       loss_from_state=False,
                       augmentor=None):
    times = []
    count = 0
    running_loss = 0.0
    if loss_fn == None:
        loss_fn = model.loss

    if rank == 0:
        aligner_validate_and_save(val_loader,
                                  model,
                                  loss_fn,
                                  mip_in,
                                  0,
                                  0,
                                  [0],
                                  reverse,
                                  checkpoint_folder=checkpoint_folder)
    for epoch in range(num_epochs):
        count = 0
        s = time.time()
        for bundle in train_loader:
            transform_seed = np.random.randint(10000000)
            np.random.seed(transform_seed)

            if augmentor is not None:
                bundle = augmentor(bundle)

            raw_bundle = copy.deepcopy(bundle)
            run_bundle = raw_bundle
            #run_bundle = apply_transform(raw_bundle, run_transform)

            run_bundle = align_sample(model, run_bundle, train=True)

            np.random.seed(transform_seed)
            raw_bundle = copy.deepcopy(bundle)
            loss_bundle = raw_bundle
            #loss_bundle = apply_transform(raw_bundle, loss_transform)
            loss_bundle['state'] = run_bundle['state']
            loss_bundle['pred_res'] = run_bundle['pred_res'].field()

            loss_bundle['pred_tgt'] = loss_bundle['pred_res'].from_pixels()(loss_bundle['src'])

            loss_dict = loss_fn(loss_bundle)
            loss_var = loss_dict['result']
            running_loss += loss_var.cpu().detach().numpy()
            if loss_var is None or loss_var != loss_var or \
                    loss_var > 1E3:
                print ("Bad loss", loss_var)
                import pdb; pdb.set_trace()
                continue

            optimizer.zero_grad()
            loss_var.backward()
            good_grad = True
            for p in model.parameters():
                if p.grad is not None and p.grad.mean() != p.grad.mean():
                    good_grad = False
                    print ("Bad grad", loss_var)
                    break
            if not good_grad:
                print ("Bad grad", loss_var)
                continue

            avg = 0
            optimizer.step()

            e = time.time()
            times.append(e - s)

            count += 1

            # if print_every is not None and ((count + 1) % print_every == 0):
            #     train_loss = running_loss / count
            #     aligner_validate_and_save(val_loader, model, loss_fn, mip_in, epoch, train_loss,
            #             times, reverse,
            #             checkpoint_folder=checkpoint_folder)

            #     running_loss = 0.0
            s = time.time()

        if (print_every is None) and (rank == 0):
            train_loss = running_loss / count
            aligner_validate_and_save(val_loader, model, loss_fn, mip_in, epoch, train_loss,
                    times, reverse,
                    checkpoint_folder=checkpoint_folder)
            times = []
            running_loss = 0.0
        # dist.barrier()
        # # synchronize processes from same checkpoint
        # map_location = {'cuda:%d' % 0: 'cuda:%d' % rank}
        # model.module.aligner.load_checkpoint(checkpoint_folder, map_location=map_location)


def aligner_validate_and_save(val_loader, model, loss_fn, mip_in, epoch, train_loss, times, reverse,
                      checkpoint_folder):
    with torch.set_grad_enabled(False):
        np.random.seed(5566)
        if not val_loader is None:
            val_loss_fn = loss_fn
            loss_dict = loss.get_dataset_loss(model, val_loader, loss_fn=val_loss_fn, mip_in=mip_in, reverse=reverse)
            val_loss = loss_dict['result']
            vec_magn = loss_dict['vec_magnitude']
            smooth_loss = loss_dict['smoothness']
            sim_loss = loss_dict['similarity']
            val_report = 'Val: {:.2E}, Vec M: {:.2E}, Sm: {:.2E}, Sim: {:.2E}'.format(val_loss, vec_magn, smooth_loss, sim_loss)
        else:
            val_report = ''

        np.random.seed()
        print ("Epoch {}: Tra: {:.2E}, {}, T: {:.4}sec".format(epoch, float(train_loss), val_report, np.mean(times)))

        sys.stdout.flush()
        model.module.aligner.save_checkpoint(checkpoint_folder=checkpoint_folder)



import time
import h5py
import json
import sys
import os
import six
import copy
import argparse

import torch
import torchfields

import numpy as np

from pathlib import Path
from tqdm import tqdm
from cloudvolume import CloudVolume

def get_dset_path(dst_folder,
                  x_offset,
                  y_offset,
                  z_start,
                  mip,
                  suffix):
    """Get H5 filepath
    """
    suffix = '_' + suffix if suffix is not None else ''
    dset_name = "field_0_x{}_y{}_z{}_MIP{}{}.h5".format(x_offset,
                                                        y_offset,
                                                        z_start,
                                                        mip,
                                                        suffix)
    return dst_folder / dset_name

def write_tensor(dset, data, sample_index, pair_index):
    """Write tensor to H5 file

    Args:
        dset (h5py.File)
        data (torch.Tensor): no leading identity dimensions
        sample_index (int)
        pair_index (int)
    """
    if data.is_cuda:
        data = data.cpu()
    data = data.numpy()
    dset[sample_index, pair_index] = data

def make_field_dset(dset_path,
                  num_samples,
                  patch_size,
                  chunk_size=512,
                  dtype=np.float32):
    """Define H5 file for data_name

    Args:
        dset_path (str): H5 filepath
        num_samples (int)
        patch_size (int): W x H; W==H for each sample
        chunk_size (int): H5 chunking (default: patch_size)
        dtype (type): datatype of H5

    Returns:
        h5py.File object, sized:
            num_samples x 2 x patch_size x patch_size
    """
    print('make_field_dset')
    if chunk_size is None:
        chunk_size = patch_size
    data_name = 'field'
    scaleoffset = 2
    df = h5py.File(dset_path, 'a')
    dset_shape = (num_samples, 2, 2, patch_size, patch_size)
    chunk_dim = (1, 1, 2, chunk_size, chunk_size)
    if data_name in df:
        del df[data_name]
    dset = df.create_dataset(data_name,
                             dset_shape,
                             dtype=dtype,
                             chunks=chunk_dim,
                             compression='lzf',
                             scaleoffset=scaleoffset)
    return dset

def make_offset_dset(dset_path,
                    num_samples,
                    dtype=int):
    """Define H5 file for field offsets

    Args:
        dset_path (str): H5 filepath
        num_samples (int)
        dtype (type): datatype of H5

    Returns:
        h5py.Dataset object, sized:
            num_samples x 2 (src, tgt) x 2 (x, y)
    """
    df = h5py.File(dset_path, 'a')
    data_name = 'offset'
    if data_name in df:
        del df[data_name]
    dset_shape = (num_samples, 2, 2)
    chunk_dim = (1, 1, 2)
    return df.create_dataset(data_name,
                             dset_shape,
                             dtype=int,
                             chunks=chunk_dim,
                             compression='lzf')

def download_section_field(vol,
                        dset,
                        offsets,
                        x_offset,
                        y_offset,
                        z_range,
                        dst_mip,
                        patch_size,
                        sample_index,
                        pair_index):
    """Download field to H5 file and collect offset adjustments

    The field will not be used to warp the img and defects. Warping is handled
    in the dataloader.

    Field is assumed to be in MIP0 displacements, and converted to
    MIP displacements. Translations are stored in MIP0 displacements.

    Args:
        vol (CloudVolume)
        dset (h5py.Dataset)
        offsets (h5py.Dataset): z x src/tgt x x/y translation
        x_offset (int): vol.mip pixels
        y_offset (int): vol.mip pixels
        z_range (slice): length one
        dst_mip (int): MIP of dataset
        patch_size (int): dst_mip pixels
        sample_index (int)
        pair_index (int): (src, tgt): (0, 1)
    """
    src_mip = vol.mip
    assert(src_mip > dst_mip)
    scale_factor = 2**(src_mip - dst_mip)
    in_field = vol[x_offset:x_offset + (patch_size // scale_factor),
                   y_offset:y_offset + (patch_size // scale_factor),
                   z_range]
    in_field = np.transpose(in_field, (2,3,0,1))
    in_field = torch.tensor(in_field).field()
    in_field = in_field.squeeze()
    out_field = in_field.up(src_mip - dst_mip)
    out_field = out_field[:, :patch_size, :patch_size]
    trans = out_field.mean_finite_vector(keepdim=True)
    trans = (trans // (2**dst_mip)) * 2**dst_mip
    offset = [int(trans[0,0,0]), int(trans[1,0,0])]
    offsets[sample_index, pair_index, :] = offset
    out_field -= trans
    out_field = out_field / (2**dst_mip)
    out_field = torch.flip(out_field, [0]) # reverse x,y components
    write_tensor(dset=dset,
                 data=out_field,
                 sample_index=sample_index,
                 pair_index=pair_index)

def download_dataset_field(cv_path,
                        dst_folder,
                        z_start,
                        z_end,
                        src_mip,
                        dst_mip,
                        x_offset=0,
                        y_offset=0,
                        patch_size=None,
                        suffix=None,
                        parallel=1):
    """Create CloudVolume & H5 file and transfer field

    Args:
        cv_path (str): CloudVolume path
        dst_folder (str): root of directory where H5 files will be stored
        offset_translations (dict): z: (x trans, y trans)
        z_start (int)
        z_end (int)
        src_mip (int): MIP level of CloudVolume field
        dst_mip (int): MIP level of output dataset
        x_offset (int): offset in MIP0 pixels; must be multiple of MIP factor
        y_offset (int): offset in MIP0 pixels; must be multiple of MIP factor
        patch_size (int): width & height of 2D region to download
        suffix (str): append to each H5 filename
        parallel (int): no. of threads for CloudVolume operations
    """
    section_ids = range(z_start, z_end)
    num_samples = len(section_ids)
    dset_path = get_dset_path(dst_folder=dst_folder,
                              x_offset=x_offset,
                              y_offset=y_offset,
                              z_start=z_start,
                              mip=dst_mip,
                              suffix=suffix)
    dset = make_field_dset(dset_path=dset_path,
                          num_samples=num_samples,
                          patch_size=patch_size)
    vol = CloudVolume(cv_path,
                      mip=src_mip,
                      fill_missing=True,
                      bounded=False,
                      progress=False,
                      parallel=parallel)
    offsets = make_offset_dset(dset_path=dset_path,
                               num_samples=num_samples,
                               dtype=int)
    x_offset //= 2**src_mip
    y_offset //= 2**src_mip
    for sample_index, z in tqdm(enumerate(section_ids)):
        for pair_index in range(2):
            z_range = slice(z, z+1) if pair_index == 0 else slice(z-1, z)
            download_section_field(vol=vol,
                                dset=dset,
                                offsets=offsets,
                                x_offset=x_offset,
                                y_offset=y_offset,
                                z_range=z_range,
                                dst_mip=dst_mip,
                                patch_size=patch_size,
                                sample_index=sample_index,
                                pair_index=pair_index)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
            description='Create MetroEM field dataset via CloudVolume')
    parser.add_argument('--src_mip',         type=int)
    parser.add_argument('--dst_mip',         type=int)
    parser.add_argument('--patch_size',  type=int)
    parser.add_argument('--x_offset',    type=int, default=0)
    parser.add_argument('--y_offset',    type=int, default=0)
    parser.add_argument(
            '--z_start',
            type=int,
            default=None,
            help='Start of source image range (target range start is z_start-1')
    parser.add_argument(
            '--z_end',
            type=int,
            default=None,
            help='End of source image range (target range end is z_end-1')
    parser.add_argument('--cv_path', type=str, default=None)
    parser.add_argument('--suffix', type=str, default=None)
    parser.add_argument('--dst_folder', type=str, default='./dataset01')
    parser.add_argument('--parallel', type=int, default=1)

    args = parser.parse_args()

    dst_folder = Path(args.dst_folder)
    dst_folder.mkdir(parents=True, exist_ok=True)

    assert(args.x_offset % 2**args.src_mip == 0)
    assert(args.y_offset % 2**args.src_mip == 0)

    download_dataset_field(cv_path=args.cv_path,
                        dst_folder=dst_folder,
                        z_start=args.z_start,
                        z_end=args.z_end,
                        src_mip=args.src_mip,
                        dst_mip=args.dst_mip,
                        x_offset=args.x_offset,
                        y_offset=args.y_offset,
                        patch_size=args.patch_size,
                        suffix=args.suffix,
                        parallel=args.parallel)



import torch
import time
import numpy as np
import six

from metroem import helpers
from metroem.loss import unsupervised_loss

def combine_pre_post(res, pre, post):
    result = post.from_pixels()(res.from_pixels()(pre.from_pixels())).pixels()
    return result

def optimize_pre_post_ups(src, tgt, initial_res, sm, lr, num_iter,
                      src_defects,
                      tgt_defects,
                      opt_params={},
                      opt_mode='adam',
                      crop=16,
                      opt_res_coarsness=0, wd=1e-3, l2=0.0,
                      gridsample_mode="bilinear",
                      normalize=True,
                      sm_keys_to_apply={},
                      mse_keys_to_apply={}):
    opti_loss = unsupervised_loss(
          smoothness_factor=sm, use_defect_mask=True,
          sm_keys_to_apply=sm_keys_to_apply,
          mse_keys_to_apply=mse_keys_to_apply
      )
    pred_res = initial_res.clone().detach().field()
    pred_res = pred_res.down(opt_res_coarsness)
    pred_res.requires_grad = False

    pre_res = torch.zeros_like(pred_res, device=pred_res.device).field().detach()
    post_res = torch.zeros_like(pre_res, device=pred_res.device).field().detach()

    prev_pre_res = torch.zeros_like(pre_res, device=pred_res.device).field().detach()
    prev_post_res = torch.zeros_like(pre_res, device=pred_res.device).field().detach()
    for t in [pre_res, post_res, prev_pre_res, prev_post_res]:
        t.requires_grad = True
    trainable = [pre_res, post_res]
    if opt_mode == 'adam':
        optimizer = torch.optim.Adam(trainable, lr=lr, weight_decay=wd)
    elif opt_mode == 'sgd':
        optimizer = torch.optim.SGD(trainable, lr=lr, **opt_params)

    loss_bundle = {
        'src': src,
        'tgt': tgt,
        'tgt_defects': tgt_defects,
    }
    loss_bundle['src_defects'] = src_defects
    prev_loss = []

    s = time.time()
    if normalize:
        with torch.no_grad():
            loss_bundle = helpers.normalize_bundle(loss_bundle, per_feature_var=True, mask_fill=0)

    loss_bundle['pred_res'] = combine_pre_post(pred_res, pre_res, post_res).up(opt_res_coarsness)

    loss_bundle['pred_tgt'] = loss_bundle['pred_res'].from_pixels()(src)
    loss_dict = opti_loss(loss_bundle, crop=crop)
    best_loss = loss_dict['result'].cpu().detach().numpy()
    new_best_ago = 0
    lr_halfed_count = 0
    nan_count = 0
    no_impr_count = 0
    new_best_count = 0
    print (loss_dict['result'].cpu().detach().numpy(), loss_dict['similarity'].detach().cpu().numpy(), loss_dict['smoothness'].detach().cpu().numpy())

    for epoch in range(num_iter):
        loss_bundle['pred_res'] = combine_pre_post(pred_res, pre_res, post_res).up(opt_res_coarsness)

        loss_bundle['pred_tgt'] = loss_bundle['pred_res'].from_pixels()(src)
        loss_dict = opti_loss(loss_bundle, crop=crop)
        loss_var = loss_dict['result']
        #print (loss_dict['result'].cpu().detach().numpy(), loss_dict['similarity'].detach().cpu().numpy(), loss_dict['smoothness'].detach().cpu().numpy())
        loss_var += (loss_bundle['pred_res']**2).mean() * l2
        curr_loss = loss_var.cpu().detach().numpy()

        #print (loss_dict['result'].cpu().detach().numpy(), loss_dict['similarity'].detach().cpu().numpy(), loss_dict['smoothness'].detach().cpu().numpy())
        if np.isnan(curr_loss):
            nan_count += 1
            lr /= 1.5
            lr_halfed_count += 1
            pre_res = prev_pre_res.clone().detach()
            post_res = prev_post_res.clone().detach()
            post_res.requires_grad = True
            pre_res.requires_grad = True
            trainable = [pre_res, post_res]
            if opt_mode == 'adam':
                optimizer = torch.optim.Adam([pre_res, post_res], lr=lr, weight_decay=wd)
            elif opt_mode == 'sgd':
                optimizer = torch.optim.SGD([pre_res, post_res], lr=lr, **opt_params)
            prev_loss = []
            new_best_ago = 0
        else:
            min_improve = 1e-11
            if not np.isnan(curr_loss) and curr_loss + min_improve <= best_loss:
                prev_pre_res = pre_res.clone()
                prev_post_res = post_res.clone()
                best_loss = curr_loss
                #print ("new best")
                new_best_count += 1
                new_best_ago = 0
            else:
                new_best_ago += 1
                if new_best_ago > 12:
                    #print ("No improvement, reducing lr")
                    no_impr_count += 1
                    lr /= 2
                    lr_halfed_count += 1
                    #pre_res = prev_pre_res.clone().detach()
                    #post_res = prev_post_res.clone().detach()
                    #post_res.requires_grad = True
                    #pre_res.requires_grad = True
                    if opt_mode == 'adam':
                        optimizer = torch.optim.Adam([pre_res, post_res], lr=lr)
                    elif opt_mode == 'sgd':
                        optimizer = torch.optim.SGD([pre_res, post_res], lr=lr, **opt_params)
                    new_best_ago -= 5
                prev_loss.append(curr_loss)

            optimizer.zero_grad()
            loss_var.backward()
            #torch.nn.utils.clip_grad_norm([pre_res, post_res], 4e0)
            pre_res.grad[pre_res.grad != pre_res.grad] = 0
            post_res.grad[post_res.grad != post_res.grad] = 0
            optimizer.step()

            if lr_halfed_count >= 15 or nan_count > 15:
                break


    loss_bundle['pred_res'] = combine_pre_post(pred_res, prev_pre_res, prev_post_res).up(opt_res_coarsness)

    loss_bundle['pred_tgt'] = loss_bundle['pred_res'].from_pixels()(src)
    loss_dict = opti_loss(loss_bundle, crop=crop)

    e = time.time()
    print ("New best: {}, No impr: {}, NaN: {}, Iter: {}".format(new_best_count, no_impr_count, nan_count, epoch))
    print (loss_dict['result'].cpu().detach().numpy(), loss_dict['similarity'].detach().cpu().numpy(), loss_dict['smoothness'].detach().cpu().numpy())
    print (e - s)
    print ('==========')


    return loss_bundle['pred_res']



import os
import numpy as np
import collections
import torch
import torch.nn as nn
import torch.nn.functional as F
from skimage.transform import rescale
from functools import reduce
import subprocess
from torch.nn.parameter import Parameter
from cloudvolume import CloudVolume

import h5py
import artificery
import json

def get_random_crop_coords(full_shape, cropped_shape, coord_granularity=4):
    assert cropped_shape[0] <= full_shape[0]
    assert cropped_shape[1] <= full_shape[1]
    assert cropped_shape[0] % coord_granularity == 0
    assert cropped_shape[1] % coord_granularity == 0

    x_bot_preshift = np.random.randint(0, full_shape[0] - cropped_shape[0] + 1)
    y_bot_preshift = np.random.randint(0, full_shape[1] - cropped_shape[1] + 1)
    x_bot = x_bot_preshift - (x_bot_preshift % coord_granularity)
    y_bot = y_bot_preshift - (y_bot_preshift % coord_granularity)
    x_top = x_bot + cropped_shape[0]
    y_top = y_bot + cropped_shape[1]

    return x_bot, x_top, y_bot, y_top

def get_center_crop_coords(full_shape, cropped_shape, coord_granularity=4):
    assert cropped_shape[0] <= full_shape[0]
    assert cropped_shape[1] <= full_shape[1]

    assert cropped_shape[0] % coord_granularity == 0
    assert cropped_shape[1] % coord_granularity == 0

    x_bot_preshift = (full_shape[0] - cropped_shape[0]) // 2
    y_bot_preshift = (full_shape[1] - cropped_shape[1]) // 2
    x_bot = x_bot_preshift - (x_bot_preshift % coord_granularity)
    y_bot = y_bot_preshift - (y_bot_preshift % coord_granularity)

    x_top = x_bot + cropped_shape[0]
    y_top = y_bot + cropped_shape[1]

    return x_bot, x_top, y_bot, y_top

def random_crop(img, cropped_shape):
    result = []
    if isinstance(img, list):
        original_shape = img[0].shape[-2:]
        x_bot, x_top, y_bot, y_top = get_random_crop_coords(original_shape, cropped_shape)
        for i in img:
            assert (i.shape[-2] == original_shape[-2])
            assert (i.shape[-1] == original_shape[-1])

            result.append(i[..., x_bot:x_top, y_bot:y_top])
    else:
        original_shape = img.shape
        x_bot, x_top, y_bot, y_top = get_random_crop_coords(original_shape, cropped_shape)

        result.append(img[..., x_bot:x_top, y_bot:y_top])
    return result


def open_model(name, checkpoint_folder):
    a = artificery.Artificery()

    spec_path = os.path.join(checkpoint_folder, "model_spec.json")
    my_p = a.parse(spec_path)

    checkpoint_path = os.path.join(checkpoint_folder, "{}.state.pth.tar".format(name))
    if os.path.isfile(checkpoint_path):
        my_p.load_state_dict(torch.load(checkpoint_path))
    my_p.name = name
    return my_p.cuda()

def create_model(name, model_spec, checkpoint_folder, write_spec=True):
    checkpoint_path = os.path.join(checkpoint_folder, "{}.state.pth.tar".format(name))
    print (checkpoint_path)
    if os.path.isfile(checkpoint_path):
        #my_p.load_state_dict(torch.load(checkpoint_path))
        a = artificery.Artificery(checkpoint_init=False)
    else:
        a = artificery.Artificery(checkpoint_init=True)

    spec_path = os.path.expanduser(model_spec["spec_path"])
    spec_dir = os.path.dirname(spec_path)
    my_p = a.parse(model_spec["spec_path"])
    my_p.name = name

    if write_spec:
        model_spec_dst = os.path.join(checkpoint_folder, "model_spec.json")
        subprocess.Popen("ls {}".format(spec_path), shell=True)
        subprocess.Popen("cp {} {}".format(spec_path, model_spec_dst), shell=True)
        for sf in a.used_specfiles:
            sf = os.path.expanduser(sf)
            rel_folder = os.path.dirname(os.path.relpath(sf, spec_dir))
            dst_folder = os.path.join(checkpoint_folder, rel_folder)
            if not os.path.exists(dst_folder):
                os.makedirs(dst_folder)
            subprocess.Popen("cp {} {}".format(sf, dst_folder), shell=True)

    if os.path.isfile(checkpoint_path):
        #my_p.load_state_dict(torch.load(checkpoint_path))
        state_dict = torch.load(checkpoint_path)

        own_state = my_p.state_dict()
        reinit_downmodules = [7]
        reinit_upmodules = []

        for layer_name, param in state_dict.items():
            load_weights = True
            if "level_downmodules" in layer_name:
                level = layer_name[len("level_downmodules.")]
                level = int(level[:1])
                if level in reinit_downmodules:
                    load_weights = False

            elif "level_upmodules" in layer_name:
                level = layer_name[len("level_upmodules.")]
                level = int(level[:1])
                if level in reinit_upmodules:
                    load_weights = False

            if layer_name not in own_state:
                load_weights = False

            if load_weights:
                if isinstance(param, Parameter):
                    # backwards compatibility for serialized parameters
                    param = param.data
                own_state[layer_name].copy_(param)
                #print ("Overwriting '{}'".format(layer_name))


    return my_p

def to_tensor(np_array, device='cuda'):
    return torch.tensor(np_array, device=device)

def expand_dims(tensor, dim_out):
    dim_in = len(list(tensor.shape))
    assert dim_out >= dim_in
    for i in range(dim_in, dim_out):
        tensor = tensor.unsqueeze(0)
    return tensor

def get_np(pt):
        return pt.cpu().detach().numpy()

def compose_functions(fseq):
    def compose(f1, f2):
        return lambda x: f2(f1(x))
    return reduce(compose, fseq, lambda _: _)


def np_upsample(img, factor):
    if factor == 1:
        return img

    if img.ndim == 2:
        return rescale(img, factor)
    elif img.ndim == 3:
        b = np.empty((int(img.shape[0] * factor),
                      int(img.shape[1] * factor), img.shape[2]))
        for idx in range(img.shape[2]):
            b[:, :, idx] = np_upsample(img[:, :, idx], factor)
        return b
    else:
        assert False


def np_downsample(img, factor):
    data_4d = np.expand_dims(img, axis=1)
    result = nn.AvgPool2d(factor)(torch.from_numpy(data_4d))
    return result.numpy()[:, 0, :, :]


def center_field(field):
    wrap = type(field) == np.ndarray
    if wrap:
        field = [field]
    for idx, vfield in enumerate(field):
        vfield[:, :, :, 0] = vfield[:,:,:,0] - np.mean(vfield[:,:,:,0])
        vfield[:, :, :, 1] = vfield[:,:,:,1] - np.mean(vfield[:,:,:,1])
        field[idx] = vfield
    return field[0] if wrap else field


def reverse_dim(var, dim):
    if var is None:
        return var
    idx = range(var.size()[dim] - 1, -1, -1)
    idx = torch.LongTensor(idx)
    if var.is_cuda:
        idx = idx.cuda()
    return var.index_select(dim, idx)


def reduce_seq(seq, f):
    size = min([x.size()[-1] for x in seq])
    return f([center(var, (-2, -1),
              var.size()[-1] - size) for var in seq], 1)


def center(var, dims, d):
    if not isinstance(d, collections.Sequence):
        d = [d for i in range(len(dims))]
    for idx, dim in enumerate(dims):
        if d[idx] == 0:
            continue
        var = var.narrow(dim, d[idx]/2, var.size()[dim] - d[idx])
    return var


def crop(data_2d, crop):
    return data_2d[..., crop:-crop, crop:-crop]


def downsample(x):
    if x > 0:
        return nn.AvgPool2d(2**x, count_include_pad=False)
    else:
        return (lambda y: y)


def upsample(x):
    if x > 0:
        return nn.Upsample(scale_factor=2**x, mode='bilinear')
    else:
        return (lambda y: y)

def normalize_bundle(bundle, per_feature_center=True,
        per_feature_var=True, mask_fill=None,
        mask_defects=True, mask_zeros=True):
    for name in ['src', 'tgt']:
        img = bundle[name]
        mask = torch.ones(img.shape[-2:],
                device=img.device, dtype=torch.uint8)
        if mask_zeros and (name + '_zeros') in bundle:
            zero_mask = bundle[name + '_zeros'].squeeze()
            mask[zero_mask != 0] = 0

        if mask_defects and (name + '_defects') in bundle:
            defect_mask = bundle[name + '_defects'].squeeze()
            mask[defect_mask != 0] = 0

        mask = mask.unsqueeze(0).unsqueeze(0).bool()
        bundle[name] = normalize(img, per_feature_center,
                        per_feature_var, mask=mask,
                        mask_fill=mask_fill)
    return bundle

def normalize(img, per_feature_center=True, per_feature_var=False, eps=1e-8,
        mask=None, mask_fill=None):
    img_out = img.clone()
    mask = mask.bool()
    #with torch.no_grad():
    if mask is not None: #black masks are shared accross featuremaps
        assert mask.shape[-1] == img.shape[-1]
    #if hasattr(mask, 'bool'):
    #    mask = mask.bool()
    mask = mask > 0
    for i in range(1):
        for b in range(img.shape[0]):
            x = img_out[b]

            if mask is not None:
                m = mask[b, 0]
            else:
                m = torch.ones_like(x[0])
            if per_feature_center and len(img.shape) == 4:
                for f in range(img.shape[1]):
                    x[f][m] = x[f][m].clone() - torch.mean(x[f][m].clone())
            else:
                x[:, m] = x[:, m].clone() - torch.mean(x[:, m].clone())
            if per_feature_var and len(img.shape) == 4:
                for f in range(img.shape[1]):
                    var = torch.var(x[f][m].clone())
                    if var == var:
                        x[f][m] = x[f][m].clone() / (torch.sqrt(var + eps))
            else:
                var = torch.var(x[:, m].clone())
                if var == var:
                    x[:, m] = x[:, m].clone() / (torch.sqrt(var + eps))

    if mask is not None and mask_fill is not None:
        for b in range(img.shape[0]):
            img_out[b, :, mask[b].squeeze() == False] = mask_fill

    return img_out




import torch
import torch.nn as nn
from torch.autograd import Variable
import numpy as np
import six
import copy
from collections import defaultdict

from pdb import set_trace as st

from metroem import helpers
from metroem.masks import get_mse_and_smoothness_masks
from metroem.alignment import align_sample

def lap(fields):
    def dx(f):
        p = Variable(torch.zeros((1,1,f.size(1),2))).cuda()
        return torch.cat((p, f[:,1:-1,:,:] - f[:,:-2,:,:], p), 1)
    def dy(f):
        p = Variable(torch.zeros((1,f.size(1),1,2))).cuda()
        return torch.cat((p, f[:,:,1:-1,:] - f[:,:,:-2,:], p), 2)
    def dxf(f):
        p = Variable(torch.zeros((1,1,f.size(1),2))).cuda()
        return torch.cat((p, f[:,1:-1,:,:] - f[:,2:,:,:], p), 1)
    def dyf(f):
        p = Variable(torch.zeros((1,f.size(1),1,2))).cuda()
        return torch.cat((p, f[:,:,1:-1,:] - f[:,:,2:,:], p), 2)
    fields = map(lambda f: [dx(f), dy(f), dxf(f), dyf(f)], fields)
    fields = map(lambda fl: (sum(fl) / 4.0) ** 2, fields)
    field = sum(map(lambda f: torch.sum(f, -1), fields))
    return field


def jacob(fields):
    def dx(f):
        p = Variable(torch.zeros((f.size(0),1,f.size(1),2))).cuda()
        return torch.cat((p, f[:,2:,:,:] - f[:,:-2,:,:], p), 1)
    def dy(f):
        p = Variable(torch.zeros((f.size(0),f.size(1),1,2))).cuda()
        return torch.cat((p, f[:,:,2:,:] - f[:,:,:-2,:], p), 2)
    fields = sum(map(lambda f: [dx(f), dy(f)], fields), [])
    field = torch.sum(torch.cat(fields, -1) ** 2, -1)
    return field


def cjacob(fields):
    def center(f):
        fmean_x, fmean_y = torch.mean(f[:,:,:,0]).data[0], torch.mean(f[:,:,:,1]).data[0]
        fmean = torch.cat((fmean_x * torch.ones((1,f.size(1), f.size(2),1)), fmean_y * torch.ones((1,f.size(1), f.size(2),1))), 3)
        fmean = Variable(fmean).cuda()
        return f - fmean

    def dx(f):
        p = Variable(torch.zeros((1,1,f.size(1),2))).cuda()
        d = torch.cat((p, f[:,2:,:,:] - f[:,:-2,:,:], p), 1)
        return center(d)
    def dy(f):
        p = Variable(torch.zeros((1,f.size(1),1,2))).cuda()
        d = torch.cat((p, f[:,:,2:,:] - f[:,:,:-2,:], p), 2)
        return center(d)

    fields = sum(map(lambda f: [dx(f), dy(f)], fields), [])
    field = torch.sum(torch.cat(fields, -1) ** 2, -1)
    return field


def njacob(fields):
    f = fields[0]
    f2 = torch.tensor(f, requires_grad=True)
    f2[:, :, :, 0] = f[:, :, :, 0] / torch.mean(torch.abs(f[:, :, :, 0]))
    f2[:, :, :, 1] = f[:, :, :, 1] / torch.mean(torch.abs(f[:, :, :, 1]))
    return jacob([f2])


def tv(fields):
    def dx(f):
        p = Variable(torch.zeros((1,1,f.size(1),2))).cuda()
        return torch.cat((p, f[:,2:,:,:] - f[:,:-2,:,:], p), 1)
    def dy(f):
        p = Variable(torch.zeros((1,f.size(1),1,2))).cuda()
        return torch.cat((p, f[:,:,2:,:] - f[:,:,:-2,:], p), 2)
    fields = sum(map(lambda f: [dx(f), dy(f)], fields), [])
    field = torch.sum(torch.abs(torch.cat(fields, -1)), -1)
    return field


def field_dx(f, forward=False):
    if forward:
        delta = f[:,1:-1,:,:] - f[:,2:,:,:]
    else:
        delta = f[:,1:-1,:,:] - f[:,:-2,:,:]
    result = delta
    result = torch.nn.functional.pad(delta, pad=(0, 0, 0, 0, 1, 1, 0, 0))
    return result

def field_dy(f, forward=False):
    if forward:
        delta = f[:,:,1:-1,:] - f[:,:,2:,:]
    else:
        delta = f[:,:,1:-1,:] - f[:,:,:-2,:]
    result = delta
    result = torch.nn.functional.pad(delta, pad=(0, 0, 1, 1, 0, 0, 0, 0))
    return result

def field_dxy(f, forward=False):
    if forward:
        delta = f[:,1:-1,1:-1,:] - f[:,2:,2:,:]
    else:
        delta = f[:,1:-1,1:-1,:] - f[:,:-2,:-2,:]

    result = delta
    result = torch.nn.functional.pad(delta, pad=(0, 0, 1, 1, 1, 1, 0, 0))
    return result

def field_dxy2(f, forward=False):
    if forward:
        delta = f[:, 1:-1, 1:-1, :] - f[:, 2:, :-2, :]
    else:
        delta = f[:, 1:-1, 1:-1, :] - f[:, :-2, 2:, :]

    result = delta
    result = torch.nn.functional.pad(delta, pad=(0, 0, 1, 1, 1, 1, 0, 0))
    return result

def rigidity_score(field_delta, tgt_length, power=2):
    spring_lengths = torch.sqrt(field_delta[..., 0]**2 + field_delta[..., 1]**2 + 1e-8)
    spring_deformations = (spring_lengths - tgt_length).abs() ** power
    return spring_deformations

def pix_identity(size, batch=1, device='cuda'):
    result = torch.zeros((batch, size, size, 2), device=device)
    x = torch.arange(size, device=device)
    result[:, :, :, 1] = x
    result = torch.transpose(result, 1, 2)
    result[:, :, :, 0] = x
    result = torch.transpose(result, 1, 2)
    return result

def rigidity(field, power=2, diagonal_mult=0.8, two_diagonals=True):
    field = field.permute(0, 2, 3, 1)
    identity = pix_identity(size=field.shape[-2], device=field.device)
    field_abs = field + identity
    result = rigidity_score(field_dx(field_abs, forward=False), 1, power=power)
    result += rigidity_score(field_dx(field_abs, forward=True), 1, power=power)
    result += rigidity_score(field_dy(field_abs, forward=False), 1, power=power)
    result += rigidity_score(field_dy(field_abs, forward=True), 1, power=power)
    result += rigidity_score(field_dxy(field_abs, forward=True), 2**(1/2), power=power) * diagonal_mult
    result += rigidity_score(field_dxy(field_abs, forward=False), 2**(1/2), power=power) * diagonal_mult
    total = 4 + 2*diagonal_mult
    if two_diagonals:
        result += rigidity_score(field_dxy2(field_abs, forward=True), 2**(1/2), power=power) * diagonal_mult
        result += rigidity_score(field_dxy2(field_abs, forward=False), 2**(1/2), power=power) * diagonal_mult
        total += 2*diagonal_mult

    result /= total

    #compensate for padding
    result[..., 0:6, :] = 0
    result[..., -6:, :] = 0
    result[..., :,  0:6] = 0
    result[..., :, -6:] = 0

    return result.squeeze()

def smoothness_penalty(ptype='jacob'):
    def penalty(fields, weights=None):
        if ptype ==     'lap': field = lap(fields)
        elif ptype == 'jacob': field = jacob(fields)
        elif ptype == 'cjacob': field = cjacob(fields)
        elif ptype == 'njacob': field = njacob(fields)
        elif ptype ==    'tv': field = tv(fields)
        elif ptype == 'rig': field = rigidity(fields[0])
        elif ptype == 'linrig': field = rigidity(fields[0], power=1)
        else: raise ValueError("Invalid penalty type: {}".format(ptype))

        #if weights is not None:
        #    field = field * weights
        return field
    return penalty


def get_dataset_loss(model, dataset_loader, loss_fn, mip_in, *args, **kwargs):
    losses = {}
    losses['result'] = []
    for sample in dataset_loader:
        aligned_bundle = align_sample(model, sample)
        if aligned_bundle is not None:
            loss_result = loss_fn(aligned_bundle)
            if isinstance(loss_result, dict):
                for k, v in six.iteritems(loss_result):
                    if isinstance(loss_result[k], torch.Tensor):
                        if k not in losses:
                            losses[k] = []
                        losses[k].append(loss_result[k].cpu().detach().numpy())
            else:
                losses['result'].append(loss_result.cpu().detach().numpy())
    for k in losses.keys():
        losses[k] = np.average(losses[k])
    return losses

def similarity_score(bundle, weights=None, crop=32):
    tgt = bundle['tgt']
    pred_tgt = bundle['pred_tgt']
    mse = ((tgt - pred_tgt)**2)
    if crop > 0:
        mse = mse[..., crop:-crop, crop:-crop]
    if weights is not None:
        weights = weights
        if crop > 0:
            weights = weights[..., crop:-crop, crop:-crop]
        total_mse = torch.sum(mse * weights)
        mask_sum  = torch.sum(weights)
        if mask_sum == 0:
            return total_mse
        else:
            return total_mse / mask_sum
    else:
        return torch.mean(mse)

def smoothness_score(bundle, smoothness_type,
                     weights=None, crop=8):
    pixelwise = smoothness_penalty(smoothness_type)([bundle['pred_res']])
    if crop > 0:
        pixelwise = pixelwise[..., crop:-crop, crop:-crop]

    if weights is not None:
        weights = weights
        if crop > 0:
            weights = weights[..., crop:-crop, crop:-crop]
        total_sm = torch.sum(pixelwise * weights)
        mask_sum  = torch.sum(weights)
        if mask_sum == 0:
            return total_sm
        else:
            return total_sm / mask_sum
    else:
        return torch.mean(pixelwise)


def similarity_sampling_loss(sample_size, unsup_loss, sample_coverage=0.001, min_nonblack=0.5):
    downsampler = torch.nn.functional.avg_pool2d

    def loss_fn(bundle):
        mse_mask = unsup_loss(bundle, smoothness_mult=0)['mse_mask']
        diff = (bundle['src'] - bundle['tgt'])**2
        mse_mask_tiles = downsampler(mse_mask, sample_size)
        diff_tiles = downsampler(diff * mse_mask, sample_size)
        result = diff_tiles / (mse_mask_tiles + 1e-8)
        result[mse_mask_tiles < min_nonblack] = 0
        #defec_tile_ids = (mse_mask_tiles > min_nonblack) * (mse_mask_tiles)
        #defect_tiles = result[result[(mse_mask_tiles > min_nonblack) * mse_mask_tiles ]]
        good_tiles = result[result != 0]
        if len(good_tiles) < 10:
            return {'similarity': 0}
        median = np.median(good_tiles.detach().cpu().numpy())
        lower_cutoff = np.percentile(good_tiles.detach().cpu().numpy(), 15)
        sampled_result = good_tiles[(good_tiles > lower_cutoff) * (good_tiles < median)].mean()
        '''tile_num = len(good_tiles)

        sample_num = int(tile_num * sample_coverage)
        if sample_num == 0:
            return {'similarity': 0}
        actual_coverage = sample_num / tile_num
        sample_result = 0
        perm = torch.randperm(good_tiles.size(0))
        idx = perm[:sample_num]
        chosen_samples = good_tiles[idx]
        sampled_result = chosen_samples.mean()
        '''
        #print ("{} -> {}".format(diff[mse_mask > 0].mean(), sampled_result))
        return {'similarity': sampled_result}


    return loss_fn


def multilevel_metric_loss(levels, mip_in, loss_fn,
        norm_embeddings=True, sm_div=False, train_thru=True,
        pre_align_sample=None):

    def compute_loss(loss_bundle, crop=32):
        loss_dict = defaultdict(lambda : 0)
        state = loss_bundle['state']


        for l in levels:
            if pre_align_sample is None:
                pre_align_loss_fn = loss_fn
            else:
                pre_align_loss_fn = similarity_sampling_loss(
                        sample_size=pre_align_sample,
                        unsup_loss=loss_fn)

            loss_bundle_emb = copy.copy(loss_bundle)
            embedding = state['up'][str(l)]['skip']

            if not train_thru:
                embedding = embedding.clone().data

            num_features = embedding.shape[1]
            loss_bundle_emb['src'] = embedding[:, 1:num_features//2]
            loss_bundle_emb['tgt'] = embedding[:, 1+num_features//2:]
            for _ in range(l - mip_in):
                for k in loss_bundle.keys():
                    if 'field' in k or 'res' in k:
                        loss_bundle_emb[k] = loss_bundle_emb[k].from_pixels().down().pixels()
                    elif 'src_' in k or 'tgt_' in k:
                        if loss_bundle_emb[k].dtype == torch.bool:
                            loss_bundle_emb[k] = torch.nn.functional.max_pool2d(loss_bundle_emb[k].float(), 2).bool()
                        else:
                            loss_bundle_emb[k] = torch.nn.functional.max_pool2d(loss_bundle_emb[k], 2)
            if norm_embeddings:
                with torch.set_grad_enabled(True):
                    loss_bundle_emb = helpers.normalize_bundle(loss_bundle_emb, per_feature_var=True, mask_fill=0)

            loss_bundle_emb['pred_tgt'] = loss_bundle_emb['src']
            #if 'src_field' in loss_bundle_emb:
            #    loss_bundle_emb['pred_tgt'] = \
            #            loss_bundle_emb['src_field'].field().from_pixels()(loss_bundle_emb['src'])


            pre_align_loss =  pre_align_loss_fn(loss_bundle_emb)['similarity']
            loss_dict['result'] = loss_dict['result'] - 0.9 * pre_align_loss / len(levels)

            if sm_div:
                sm_mult = 1.0 / 2**(l - mip_in)
            else:
                sm_mult = 1.0

            loss_bundle_emb['pred_tgt'] = loss_bundle_emb['pred_res'].from_pixels()(loss_bundle_emb['src'])
            level_loss_dict = loss_fn(loss_bundle_emb, smoothness_mult=sm_mult, crop=crop)
            loss_dict['result'] = loss_dict['result'] + level_loss_dict['result'] / len(levels)
            loss_dict['similarity'] = loss_dict['similarity'] + level_loss_dict['similarity'] / len(levels)
            if loss_dict['vec_magnitude'] == 0:
                loss_dict['vec_magnitude'] = level_loss_dict['vec_magnitude']

            loss_dict['smoothness'] = loss_dict['smoothness'] + level_loss_dict['smoothness'] / len(levels)
            #print ("Level {} posst alignment: sim {:.2E}, sm {:.2E}".format(l, level_loss_dict['similarity'], float(level_loss_dict['smoothness'])))
            #print ("Mean: {:.2E}, Var: {:.2E}".format(
            #loss_bundle_emb['src'].mean(),
            #loss_bundle_emb['src'].var()
            #    ))
        return loss_dict

    return compute_loss

def unsupervised_loss(smoothness_factor, smoothness_type='rig', use_defect_mask=False,
                      sm_keys_to_apply={}, mse_keys_to_apply={}):
    def compute_loss(bundle, smoothness_mult=1.0, crop=32):
        loss_dict = {}
        if use_defect_mask:
            mse_mask, smoothness_mask = get_mse_and_smoothness_masks(bundle,
                    sm_keys_to_apply=sm_keys_to_apply,
                    mse_keys_to_apply=mse_keys_to_apply)
        else:
            mse_mask = None
            smoothness_mask = None
        similarity = similarity_score(bundle,
                                      weights=mse_mask,
                                      crop=crop)
        if smoothness_mult != 0:
            smoothness = smoothness_score(bundle,
                                      weights=smoothness_mask,
                                      smoothness_type=smoothness_type,
                                      crop=crop)
        else:
            smoothness = torch.zeros(1, device=bundle['src'].device, dtype=torch.float32)
        result =  similarity + smoothness * smoothness_factor

        loss_dict['result'] = result
        loss_dict['similarity'] = similarity
        loss_dict['smoothness'] = smoothness * smoothness_factor * smoothness_mult
        loss_dict['vec_magnitude'] = torch.mean(torch.abs(bundle['pred_res']))
        loss_dict['vec_sim'] = torch.cuda.FloatTensor([0])
        if 'res' in bundle:
            loss_dict['vec_sim'] = torch.mean(torch.abs(bundle['pred_res'] - bundle['res']))
        loss_dict['mse_mask'] = mse_mask
        loss_dict['smoothness_mask'] = smoothness_mask
        return loss_dict
    return compute_loss





import time
import h5py
import json
import sys
import os
import six
import copy
import argparse

import torch
import torchfields

from cloudvolume import CloudVolume
import numpy as np

from pathlib import Path
from tqdm import tqdm

def get_dset_path(dst_folder,
                  x_offset,
                  y_offset,
                  z_start,
                  mip,
                  suffix):
    """Get H5 filepath
    """
    suffix = '_' + suffix if suffix is not None else ''
    dset_name = "x{}_y{}_z{}_MIP{}{}.h5".format(x_offset,
            y_offset, z_start, mip, suffix)
    return dst_folder / dset_name

def write_array(dset, data, sample_index, pair_index):
    """Write ndarray to H5 file
    """
    dset[sample_index, pair_index] = data

def make_image_dset(dset_path,
                  num_samples,
                  patch_size,
                  chunk_size=512,
                  dtype=np.float32):
    """Define H5 file for image data

    Args:
        dset_path (str): H5 filepath
        num_samples (int)
        patch_size (int): W x H; W==H for each sample
        chunk_size (int): H5 chunking (default: patch_size)
        dtype (type): datatype of H5

    Returns:
        h5py.File object, sized:
            num_samples x 2 x patch_size x patch_size
    """
    print('make_field_dset')
    if chunk_size is None:
        chunk_size = patch_size
    data_name ='img'
    data_shape = [patch_size, patch_size]
    chunk_shape = [chunk_size, chunk_size]
    df = h5py.File(dset_path, 'a')
    dset_shape = (num_samples, 2, *data_shape)
    chunk_dim = (1, 1, *chunk_shape)
    if data_name in df:
        del df[data_name]
    dset = df.create_dataset(data_name,
                             dset_shape,
                             dtype=dtype,
                             chunks=chunk_dim,
                             compression='lzf',
                             scaleoffset=None)
    return dset

def download_section_image(vol,
                        dset,
                        x_offset,
                        y_offset,
                        z_range,
                        patch_size,
                        sample_index,
                        pair_index,
                        defect_mask,
                        mask_val):
    """Download image data with defects masked to H5 files

    Args:
        vol (CloudVolume)
        dset (h5py.File)
        x_offset (int)
        y_offset (int)
        z_range (slice): length one
        patch_size (int)
        sample_index (int)
        pair_index (int): (src, tgt): (0, 1)
        defect_mask (CloudVolume)
        mask_val (number): value of mask to exclude from image
    """
    img = vol[x_offset:x_offset + patch_size,
              y_offset:y_offset + patch_size,
              z_range].squeeze((2,3))
    if defect_mask is not None:
        mask = defect_mask[x_offset:x_offset + patch_size,
                  y_offset:y_offset + patch_size,
                  z_range].squeeze((2,3))
        img[mask >= mask_val] = 0
    write_array(dset=dset,
                data=img,
                sample_index=sample_index,
                pair_index=pair_index)

def download_dataset_image(cv_path,
                        dst_folder,
                        z_start,
                        z_end,
                        mip,
                        x_offset=0,
                        y_offset=0,
                        patch_size=None,
                        suffix=None,
                        parallel=1,
                        offsets=None,
                        cv_path_defects=None,
                        mask_val=1):
    """Create CloudVolume & H5 file and transfer image

    Args:
        cv_path (str): CloudVolume path
        dst_folder (str): root of directory where H5 files will be stored
        offset_translations (dict): z: (x trans, y trans)
        z_start (int)
        z_end (int)
        mip (int)
        x_offset (int): offset in MIP0 pixels; must be multiple of MIP factor
        y_offset (int): offset in MIP0 pixels; must be multiple of MIP factor
        patch_size (int): width & height of 2D region to download
        suffix (str): append to each H5 filename
        parallel (int): no. of threads for CloudVolume operations
        offsets (h5py.Dataset): N x 2 x 2 dataset
        cv_path_defects (str): CloudVolume path to defect mask
        mask_val
    """
    section_ids = range(z_start, z_end)
    num_samples = len(section_ids)
    dset_path = get_dset_path(dst_folder=dst_folder,
                              x_offset=x_offset,
                              y_offset=y_offset,
                              z_start=z_start,
                              mip=mip,
                              suffix=suffix)
    dset = make_image_dset(dset_path=dset_path,
                          num_samples=num_samples,
                          patch_size=patch_size)
    vol = CloudVolume(cv_path,
                      mip=mip,
                      fill_missing=True,
                      bounded=False,
                      progress=False,
                      parallel=parallel)
    defect_mask = None
    if cv_path_defects is not None:
        defect_mask = CloudVolume(cv_path_defects,
                        mip=mip,
                        fill_missing=True,
                        bounded=False,
                        progress=False,
                        parallel=parallel)
    for sample_index, z in tqdm(enumerate(section_ids)):
        for pair_index in range(2):
            x_trans, y_trans = 0, 0
            if (offsets is not None):
                x_trans, y_trans = offsets[sample_index, pair_index, :]
            section_x_offset = (x_offset + x_trans) // 2**mip
            section_y_offset = (y_offset + y_trans) // 2**mip
            z_range = slice(z, z+1) if pair_index == 0 else slice(z-1, z)
            download_section_image(vol=vol,
                                dset=dset,
                                x_offset=section_x_offset,
                                y_offset=section_y_offset,
                                z_range=z_range,
                                patch_size=patch_size,
                                sample_index=sample_index,
                                pair_index=pair_index,
                                defect_mask=defect_mask,
                                mask_val=mask_val)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
            description='Create MetroEM image dataset via CloudVolume')
    parser.add_argument('--mips',        type=int, nargs='+')
    parser.add_argument('--patch_sizes', type=int, nargs='+')
    parser.add_argument('--x_offset',  type=int, default=0)
    parser.add_argument('--y_offset',  type=int, default=0)
    parser.add_argument(
            '--z_start',
            type=int,
            default=None,
            help='Start of source image range (target range start is z_start-1')
    parser.add_argument(
            '--z_end',
            type=int,
            default=None,
            help='End of source image range (target range end is z_end-1')
    parser.add_argument('--cv_path', type=str, default=None)
    parser.add_argument('--cv_path_defects', type=str, default=None)
    parser.add_argument('--mask_val', type=float, default=1)
    parser.add_argument('--field_dset',
                        type=str,
                        default=None,
                        help='Path to field dset with offsets')
    parser.add_argument('--suffix', type=str, default=None)
    parser.add_argument('--dst_folder', type=str, default='./dataset01')
    parser.add_argument('--parallel', type=int, default=1)

    args = parser.parse_args()
    assert(len(args.mips) == len(args.patch_sizes))
    dst_folder = Path(args.dst_folder)
    dst_folder.mkdir(parents=True, exist_ok=True)

    offsets = None
    if args.field_dset is not None:
        offsets = h5py.File(args.field_dset, 'r')['offset']

    for mip, patch_size in zip(args.mips, args.patch_sizes):
        download_dataset_image(cv_path=args.cv_path,
                                dst_folder=dst_folder,
                                z_start=args.z_start,
                                z_end=args.z_end,
                                mip=mip,
                                x_offset=args.x_offset,
                                y_offset=args.y_offset,
                                patch_size=patch_size,
                                suffix=args.suffix,
                                parallel=args.parallel,
                                offsets=offsets,
                                cv_path_defects=args.cv_path_defects,
                                mask_val=args.mask_val)



import sys
import json
import os
import time
import six
import argparse
import glob
import re

import modelhouse
import torch

from metroem import loss, augmentations
from metroem.alignment import aligner_train_loop
from metroem.dataset import MultimipDataset

import torch.multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist

from pdb import set_trace as st

def get_pyramid_modules(pyramid_path):
    module_dict = {}
    contents = next(os.walk(pyramid_path))[1]

    for m in contents:
        pattern = r"(\d)_mip(\d)in_(.*)"
        match = re.search(pattern, m)
        if match is None:
            raise Exception(f"Invalid member in the pyramid: {m}.\n"\
                    "Pyramid folder must only contain model folders with names in " \
                    "format '{0-9}_mip{0-9}in_{arbitrary name}'")
        module_id = int(match.group(1))
        module_mip_in = int(match.group(2))
        module_path = os.path.join(pyramid_path, m)
        module_dict[module_id] = {"path": module_path, "mip_in": module_mip_in}

    return module_dict

def setup(rank, world_size):
    dist.init_process_group(
            backend='nccl',
            init_method='env://',
            world_size=world_size,
            rank=rank)

def cleanup():
    dist.destroy_process_group()

def train_module(rank,
                 world_size,
                 module_path,
                 train_params,
                 dataset_path,
                 module_mip,
                 stage,
                 checkpoint_name,
                 aug_params=None):
    """Train object with its own dataset (specific MIP)

    Args:
        module_path (str): path to modelhouse directory
        train_params (dict)
        train_dset (AlignmentDataLoader)
        val_dset (AlignmentDataLoader)
        checkpoint_name (str)
        rank (int): process, dictates the GPU used in multi-gpu training
        world_size (int): total no. of processes
    """
    assert aug_params is None
    print(f"Running DDP on rank {rank}.")
    setup(rank, world_size)
    torch.cuda.set_device(rank)
    model = modelhouse.load_model_simple(module_path,
                                         finetune=False,
                                         pass_field=True,
                                         checkpoint_name=checkpoint_name)
    checkpoint_path = os.path.join(module_path, "model")
    model.aligner.net = model.aligner.net.to(rank)
    model = DDP(model, device_ids=[rank])

    dataset = MultimipDataset(dataset_path, aug_params, field_tag=checkpoint_name)
    train_dset = dataset.get_train_dset(mip=module_mip, stage=stage)
    val_dset = dataset.get_val_dset(mip=module_mip, stage=stage)
    val_data_loader = torch.utils.data.DataLoader(val_dset,
                                                  batch_size=1,
                                                  shuffle=True,
                                                  num_workers=0,
                                                  pin_memory=False)

    trainable = []
    trainable.extend(list(model.parameters()))

    for epoch_params in train_params:
        smoothness = epoch_params["smoothness"]
        if "print_every" in epoch_params:
            print_every = epoch_params["print_every"]
        else:
            print_every = None

        if "num_sample" in epoch_params:
            num_samples = epoch_params["num_samples"]
        else:
            num_samples = 10000000

        lr = epoch_params["lr"]
        num_epochs = epoch_params["num_epochs"]
        mse_keys_to_apply = epoch_params["mse_keys_to_apply"]
        sm_keys_to_apply = epoch_params["sm_keys_to_apply"]
        loss_spec = epoch_params["loss_spec"]
        loss_type = epoch_params["loss_spec"]["type"]

        simple_loss = loss.unsupervised_loss(smoothness,
                                             use_defect_mask=True,
                                             sm_keys_to_apply=sm_keys_to_apply,
                                             mse_keys_to_apply=mse_keys_to_apply)
        if loss_type == "plain":
            training_loss = simple_loss
        elif loss_type == "metric":
            training_loss = loss.multilevel_metric_loss(
                    loss_fn=simple_loss,
                    mip_in=0,
                    **loss_spec['params']
                    )

        augmentor = None
        if "augmentations" in epoch_params:
            augmentor = augmentations.Augmentor(epoch_params["augmentations"])

        train_dset.set_size_limit(num_samples)
        # Divide dataset across processes
        train_sampler = torch.utils.data.distributed.DistributedSampler(
                                train_dset,
                                num_replicas=world_size,
                                rank=rank)
        train_data_loader = torch.utils.data.DataLoader(train_dset,
                                                        batch_size=1,
                                                        shuffle=False,
                                                        num_workers=0,
                                                        pin_memory=False,
                                                        sampler=train_sampler)

        optimizer = torch.optim.Adam(trainable, lr=lr, weight_decay=0)
        aligner_train_loop(rank,
                           model,
                           mip_in=0,
                           train_loader=train_data_loader,
                           val_loader=val_data_loader,
                           optimizer=optimizer,
                           num_epochs=num_epochs,
                           loss_fn=training_loss,
                           print_every=print_every,
                           checkpoint_folder=checkpoint_path,
                           augmentor=augmentor)
    cleanup()
    pass

def train_pyramid(world_size,
                  pyramid_path,
                  dataset_path,
                  train_stages,
                  checkpoint_name,
                  generate_field_stages,
                  train_params=None,
                  aug_params=None):
    pyramid_path = os.path.expanduser(pyramid_path)
    module_dict = get_pyramid_modules(pyramid_path)
    print (f"Loading dataset {dataset_path}...")

    prev_mip = None
    for stage in sorted(module_dict.keys()):
        module_mip = module_dict[stage]["mip_in"]
        module_path = module_dict[stage]["path"]
        model = None

        if train_stages is None or stage in train_stages:
            print (f"Training module {stage}...")
            mip_train_params = train_params[str(module_mip)]
            mp.spawn(train_module,
                     args=(world_size,
                           module_path,
                           mip_train_params, # train_params
                           dataset_path, # dataset_path
                           module_mip, # module_mip
                           stage, # stage
                           checkpoint_name, # checkpoint_name
                           None), # aug_params
                     nprocs=world_size,
                     join=True)
           #  train_module(model, train_params=mip_train_params,
           #          train_dset=dataset.get_train_dset(mip=module_mip, stage=stage),
           #          val_dset=dataset.get_val_dset(mip=module_mip, stage=stage),
           #          checkpoint_path=os.path.join(module_path, "model"))

            print (f"Done training module {stage}!")

        if generate_field_stages is None or stage in generate_field_stages:
            print (f"Generating fields with module {stage}...")
            model = modelhouse.load_model_simple(module_path,
                    finetune=True,
                    pass_field=True,
                    finetune_iter=300//(2**stage),
                    checkpoint_name=checkpoint_name)
            dataset.generate_fields(model, mip=module_mip, stage=stage)
            print (f"Done generating fields with module {stage}...")


def main():
    parser = argparse.ArgumentParser(description='Train SeamLESS.')

    parser.add_argument('--pyramid_path', type=str)
    parser.add_argument('--dataset_path', type=str)
    parser.add_argument('--checkpoint_name', type=str, default="checkpoint")
    parser.add_argument('--train_mode', choices=['scratch', 'finetune'], type=str.lower, default='scratch')
    parser.add_argument('--train_params_path', type=str, default=None)
    parser.add_argument('--gpu', type=str, default="0")
    parser.add_argument('--train_stages', type=int, default=None, nargs='+')
    parser.add_argument('--generate_field_stages', type=int, default=None, nargs='+')
    parser.add_argument('--port', type=int, default=8888)

    parser.add_argument('--no_redirect_stdout', dest='redirect_stdout',
                        action='store_false')
    parser.set_defaults(redirect_stdout=True)
    args = parser.parse_args()

    assert(os.path.exists(args.pyramid_path))
    assert(os.path.exists(args.dataset_path))

    if args.train_mode == 'custom':
        params_path = args.train_params_path
    else:
        assert args.train_params_path is None
        dir_path = os.path.dirname(os.path.realpath(__file__))
        params_path = os.path.join(dir_path, f"params/{args.train_mode}.json")

    with open(params_path) as f:
        train_params = json.load(f)

    world_size = len(args.gpu.split(','))
    os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"]= args.gpu

    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = str(args.port)

    if args.redirect_stdout:
        log_path = os.path.join(args.pyramid_path, f"{args.checkpoint_name}.log")
        log_file = open(log_path, 'a')
        sys.stdout = log_file

    train_pyramid(world_size=world_size,
                  pyramid_path=args.pyramid_path,
                  dataset_path=args.dataset_path,
                  train_stages=args.train_stages,
                  generate_field_stages=args.generate_field_stages,
                  train_params=train_params,
                  checkpoint_name=args.checkpoint_name)

if __name__ == "__main__":
    main()



import torch

import itertools
import copy

import numpy as np

def get_peak(x, eps=1.E-3, randomize=False):
    # x_in is shape BxXxY
    d = x.shape[-1]
    peak = torch.zeros((x.shape[0], 2), device=x.device, dtype=torch.long)

    for b in range(x.shape[0]):
        m = x[b].argmax()
        initial_peak = [int(m // d), int(m % d)]

        if randomize:
            peak_value = x[b, initial_peak[0], initial_peak[1]].squeeze(0)
            peaks_mask = x[b] >= (peak_value - eps)
            peaks_coords = peaks_mask.nonzero()
            random_peak_id = np.random.randint(0, peaks_coords.shape[0])
            chosen_one = peaks_coords[random_peak_id]
            peak[b, 0] = chosen_one[0]
            peak[b, 1] = chosen_one[1]
        else:
            peak[b, 0] = initial_peak[0]
            peak[b, 1] = initial_peak[1]
    return peak

def zone_out_around_peak(x, peak, zone_dist, fill=None):
    if fill == None:
        fill = float("-inf")
    d = x.shape[-1]
    zoned_x = x.clone()
    for b in range(x.shape[0]):
        x_low = max(peak[b, 0] - zone_dist, 0)
        x_high = min(peak[b, 0] + zone_dist + 1, d)
        y_low = max(peak[b, 1] - zone_dist, 0)
        y_high = min(peak[b, 1] + zone_dist + 1, d)
        zoned_x[b, x_low:x_high, y_low:y_high] = fill
    return zoned_x

def get_two_peaks(x_in, zone_dist=1, eps=1.E-3):
    # x_in is shape BxXxY
    first_peak = get_peak(x_in)
    zoned_x = zone_out_around_peak(x_in, first_peak, zone_dist)
    second_peak = get_peak(zoned_x)
    return first_peak, second_peak


def get_black_mask(img, black_threshold):
    if black_threshold == 0:
        black_mask = img == 0
    else:
        black_mask = img <= black_threshold
    return black_mask

def get_black_fraction(img, black_threshold):
    img_black_px = get_black_mask(img, black_threshold)
    img_black_px_count = torch.sum(img_black_px)
    img_px_count = torch.sum(torch.ones_like(img))
    img_black_fraction = float(img_black_px_count) / float(img_px_count)

    return img_black_fraction

def normalize(img, per_feature_center=True, per_feature_var=False, eps=1e-5,
        mask=None, mask_fill=None):
    img_out = img.clone()
    #with torch.no_grad():
    if mask is not None:
        assert mask.shape == img.shape
    for i in range(1):
        for b in range(img.shape[0]):
            x = img_out[b]
            if per_feature_center and len(img.shape) == 4:
                for f in range(img.shape[1]):
                    if mask is not None:
                        m = mask[b, f]
                        x[f][m] = x[f][m].clone() - torch.mean(x[f][m].clone())
                    else:
                        x[f] = x[f].clone() - torch.mean(x[f].clone())
            else:
                if mask is not None:
                    m = mask[b]
                    x[m] = x[m].clone() - torch.mean(x[m].clone())
                else:
                    x[...] = x.clone() - torch.mean(x.clone())

            if per_feature_var and len(img.shape) == 4:
                for f in range(img.shape[1]):
                    if mask is not None:
                        m = mask[b, f]
                        var = torch.var(x[f][m].clone())
                        x[f][m] = x[f][m].clone() / (torch.sqrt(var) + eps)
                    else:
                        var = torch.var(x[f].clone())
                        x[f] = x[f].clone() / (torch.sqrt(var) + eps)
            else:
                if mask is not None:
                    m = mask[b]
                    var = torch.var(x[m].clone())
                    x[m] = x[m].clone() / (torch.sqrt(var) + eps)
                else:
                    var = torch.var(x.clone())
                    x[...] = x.clone() / (torch.sqrt(var) + eps)

    if mask is not None and mask_fill is not None:
        img_out[mask == False] = mask_fill

    return img_out

def get_index_neighbors(index, shape, diagonals=False, reach=1):
    result = []

    if diagonals:
        offsets = itertools.product(np.arange(0, reach + 2) - 1, repeat=2)
    else:
        offsets = []
        offsets += itertools.product(np.arange(0, reach + 2) - 1, [0])
        offsets += itertools.product([0], np.arange(0, reach + 2) - 1)

    for o in offsets:
        new_index = list(copy.copy(index))
        new_index[0] += o[0]
        new_index[1] += o[1]
        if new_index[0] >= 0 and new_index[1] >=0 and \
           new_index[0] < shape[0] and new_index[1] < shape[1]:
               result.append(new_index)
    return result

def get_neighbor_average(index, field):
    neighbors = get_index_neighbors(index, field.shape)
    result = 0
    valid_count = 0
    nonzero_count = 0
    invalid_count = 0

    for i in neighbors:
        value = field[i[0], i[1]]
        if not np.isnan(value).any() and not np.isinf(value).any():
            valid_count += 1
            if value[0] != 0 or value[1] != 0:
                result += value
                nonzero_count += 1
        else:
            invalid_count += 1

    if valid_count == 0 or (invalid_count > 0 and nonzero_count == 0):
        return None
    elif nonzero_count > 0:
        return result / nonzero_count
    else:
        return 0

def extrapolate_field_missing_values(field):
    field[..., 0] = extrapolate_missing_values(field[..., 0])
    field[..., 1] = extrapolate_missing_values(field[..., 1])
    return field

def extrapolate_missing_values(array):
    missing_map = (np.isnan(array) + np.isinf(array)) > 0
    missing_indexes = np.nonzero(missing_map)
    missing_coords = list(set(zip(missing_indexes[0], missing_indexes[1])))

    missing_count = len(missing_coords)
    while missing_count > 0:
        print ("Purging ", len(missing_coords))
        for i in missing_coords:
            v = get_neighbor_average(i, array)
            if v is not None:
                array[i[0], i[1]] = v

        missing_map = (np.isnan(array) + np.isinf(array)) > 0
        missing_indexes = np.nonzero(missing_map)
        missing_coords = list(set(zip(missing_indexes[0], missing_indexes[1])))
        prev_missing_count = missing_count
        missing_count = len(missing_coords)
        if missing_count >= prev_missing_count:
            print ("Purge stuck, 0 everything")
            break
    array[missing_indexes] = 0

    return array

class BlockMatcher:
    def __init__(self, tile_size=32, tile_step=16, max_disp=32,
            do_upsampling=True, black_threshold=0):
        self.tile_size = tile_size
        self.tile_step = tile_step
        self.max_disp = max_disp
        self.do_upsampling = do_upsampling
        self.black_threshold = black_threshold

    def forward(self, x=None, src=None, tgt=None, tile_size=None,
            tile_step=None, max_disp=None, do_upsampling=None,
            black_threshold=None,
            **kwargs):
        if src is None or tgt is None:
            assert x is not None
            src = x[:, 0]
            tgt = x[:, 1]

        if tile_size is None:
            tile_size = self.tile_size
        if tile_step is None:
            tile_step = self.tile_step
        if max_disp is None:
            max_disp = self.max_disp
        if do_upsampling is None:
            do_upsampling = self.do_upsampling
        if black_threshold is None:
            black_threshold = self.black_threshold

        result = block_match(src, tgt,
                tile_size=tile_size, tile_step=tile_step,
                max_disp=max_disp,
                peak_ratio_cutoff=0.0, min_overlap_px=200,
                do_upsampling=do_upsampling,
                black_threshold=black_threshold)
        return result

def block_match(tgt, src, tile_size=16, tile_step=16, max_disp=10, min_overlap_px=500,
                filler="inf", peak_ratio_cutoff=1.1, peak_distance=2,
                do_upsampling=False, black_threshold=0):
    src = src.squeeze()
    tgt = tgt.squeeze()
    tile_alignment_pad = (tile_size - tile_step) // 2

    padded_tgt = torch.nn.ZeroPad2d(tile_alignment_pad)(tgt)
    padded_src = torch.nn.ZeroPad2d(tile_alignment_pad)(src)

    max_disp_pad = max_disp
    padded_tgt = torch.nn.ZeroPad2d(max_disp_pad)(tgt)
    padded_src = torch.nn.ZeroPad2d(max_disp_pad)(src)

    img_size = padded_tgt.shape[-1]
    tile_count = 1 + (img_size - max_disp*2 - tile_size) // tile_step
    result = np.zeros((tile_count, tile_count, 2))
    peaks = []
    peak_vals = []
    peak_ratios = []
    for x_tile in range(0, tile_count):
        for y_tile in range(0, tile_count):
            src_tile_coord, tgt_tile_coord = compute_tile_coords(x_tile, y_tile, tile_size,
                                                              tile_step, max_disp, img_size,
                                                                x_offset=max_disp,
                                                                y_offset=max_disp)
            src_tile = padded_src[src_tile_coord]
            tgt_tile = padded_tgt[tgt_tile_coord]
            if get_black_fraction(src_tile, black_threshold) > 0.8 \
                or get_black_fraction(tgt_tile, black_threshold) > 0.9:
                match_displacement = [0, 0]
                if get_black_fraction(src_tile, black_threshold) != 1.0 and \
                        get_black_fraction(tgt_tile, black_threshold) != 1.0:
                    #print ('skipping: {} {}'.format(get_black_fraction(src_tile, 0), get_black_fraction(tgt_tile, 0)))
                    pass
            else:
                ncc = get_ncc(tgt_tile, src_tile, div_by_overlap=True,
                        min_overlap_count=min_overlap_px,
                        black_threshold=black_threshold)
                ncc_np = ncc.squeeze().cpu().numpy()

                if ncc.var() < 1E-15 or ((ncc != ncc).sum() > 0):
                    #match_displacement = [0, 0]
                    #match_displacement = [float(filler), float(filler)]
                    match_displacement = [0, 0]
                    #print ("black ncc")
                else:
                    peak1, peak2 = get_two_peaks(ncc, peak_distance)
                    peaks.append([peak1, peak2])
                    match = np.unravel_index(ncc_np.argmax(), ncc_np.shape)
                    peak_vals.append([ncc_np[peak1[0][0],  peak1[0][1]],
                                      ncc_np[peak2[0][0],  peak2[0][1]],
                                    ])
                    peak_ratios.append(peak_vals[-1][0] / (peak_vals[-1][1] + 1e-12))
                    if peak_ratios[-1] < peak_ratio_cutoff:
                        match_displacement = [float(filler), float(filler)]
                    else:
                        match_tile_start = (tgt_tile_coord[0].start + match[0], tgt_tile_coord[1].start + match[1])
                        src_tile_start   = (src_tile_coord[0].start, src_tile_coord[1].start)
                        match_displacement = np.subtract(src_tile_start, match_tile_start)

            result[x_tile, y_tile, 0] = -match_displacement[1]
            result[x_tile, y_tile, 1] = -match_displacement[0]

    #patched_result = extrapolate_missing_values(result)
    patched_result = result
    result_var = torch.cuda.FloatTensor(patched_result).unsqueeze(0)
    if do_upsampling:
        scale = tgt.shape[-2] / result_var.shape[-2]
        result_var = torch.nn.functional.interpolate(result_var.permute(0, 3, 1, 2),
            scale_factor=scale, mode='bicubic').permute(0, 2, 3, 1)

    final_result_var = result_var
    #final_result_var = filter_black_field(final_result_var, src, black_threshold=0.1)

    final_result = final_result_var
    return final_result

def filter_black_field(field, img, black_threshold=0):
    black_mask = (img.abs() < black_threshold).squeeze()
    field[..., black_mask] = 0
    return field

def get_ncc(tgt, tmpl, div_by_overlap=False, min_overlap_ratio=0.2,
            min_overlap_count=500, black_threshold=0):
    tgt = tgt.unsqueeze(0).unsqueeze(0)
    tmpl = tmpl.unsqueeze(0).unsqueeze(0)

    if black_threshold == 0:
        tgt_mask = tgt.abs() > black_threshold
        tmpl_mask = tmpl.abs() > black_threshold
    else:
        tgt_mask = tgt > black_threshold
        tmpl_mask = tmpl > black_threshold
    tgt_norm = normalize(tgt, mask=tgt_mask)
    tmpl_norm = normalize(tmpl, mask=tmpl_mask)
    ncc = get_cc(tgt_norm, tmpl_norm)

    overlap_count = get_cc(tgt_mask.float(), tmpl_mask.float(), div_by_tmpl_size=False)

    if div_by_overlap:
        overlap_ratio = overlap_count / (tmpl_mask != 0).sum()
        adjusted_ncc = ncc / (overlap_count + 1e-14)
        ncc[overlap_count != 0] = adjusted_ncc[overlap_count != 0]
        ncc[overlap_count < min_overlap_count] = ncc.min()
        ncc[overlap_ratio < min_overlap_ratio] = ncc.min()
        #print ("Bad overlap count: {}".format((overlap_ratio < min_overlap_ratio).sum()))

    return ncc

def get_cc(target, template, feature_weights=None, normalize_cc=False, div_by_tmpl_size=True):
    cc_side = target.shape[-1] - template.shape[-1] + 1
    cc = torch.zeros((target.shape[0], cc_side, cc_side),
            device=target.device, dtype=torch.float)
    for b in range(target.shape[0]):
        cc[b:b+1] = torch.nn.functional.conv2d(target[b:b+1], template[b:b+1]).squeeze()
    #print (template.shape)
    if normalize_cc:
        cc = normalize(cc)
    elif div_by_tmpl_size:
        cc = cc / torch.sum(torch.ones_like(template[0], device=template.device))
    #print (torch.mean(cc))
    return cc

def get_displaced_tile(disp, tile):
    result = copy.deepcopy(tile)
    result[0].start += disp[0]

def compute_tile_coords(x_tile, y_tile, tile_size, tile_step, max_disp, img_size,
                       x_offset=0, y_offset=0):
    #import pdb; pdb.set_trace()
    src_xs = x_tile * tile_step + x_offset
    src_xe = src_xs + tile_size
    src_ys = y_tile * tile_step + y_offset
    src_ye = src_ys + tile_size

    tgt_xs = max(0, src_xs - max_disp)
    tgt_xe = min(img_size, src_xe + max_disp)
    tgt_ys = max(0, src_ys - max_disp)
    tgt_ye = min(img_size, src_ye + max_disp)

    src_coords = (slice(src_xs, src_xe), slice(src_ys, src_ye))
    tgt_coords = (slice(tgt_xs, tgt_xe), slice(tgt_ys, tgt_ye))

    return src_coords, tgt_coords

def get_patch_middle(coords):
    x_m = (coords[0].start - coords[0].end) / 2
    y_m = (coords[1].start - coords[1].end) / 2

    return(x_m, y_m)



import torch
import torchvision

import numpy as np
import skimage

import six
import copy


class Augmentor():
    def __init__(self, augmentations=[]):
        transform = []
        for a_spec in augmentations:
            t = a_spec['type']
            del a_spec['type']

            if t == "warp":
                transform.append(RandomWarp(**a_spec))
            elif t =="random_transpose":
                transform.append(RandomTranspose(**a_spec))
            elif t =="random_src_tgt_swap":
                transform.append(RandomSrcTgtSwap(**a_spec))


        self.transform = torchvision.transforms.Compose(transform)

    def __call__(self, bundle):
        return self.transform(bundle)

class RandomWarp(object):
    """ Warp With Random Field
    difficulty
        0 == one direction
        1 == 2 directions along each axis
        i == 2^i directions along each axis
    """
    def __init__(self, difficulty=2, max_disp=10, prob=1.0, min_disp=0,
                 randomize_d=False, random_epicenters=True):
        assert difficulty >= 0
        self.randomize_d = randomize_d
        self.difficulty = difficulty
        self.max_disp = max_disp
        self.min_disp = min_disp
        self.prob     = prob
        self.random_epicenters = random_epicenters

    def __call__(self, bundle):
        src = bundle['src']

        if self.randomize_d and self.difficulty > 0:
            curr_diff = random.randint(0, self.difficulty)
        else:
            curr_diff = self.difficulty
        coin = np.random.uniform()
        if coin < self.prob:
            # NOTE image dimention assumed to be power of 2

            granularity = int(np.log2(src.shape[-1]) - curr_diff)
            res_delta = generate_random_residuals(src.squeeze().shape,
                                                  min_disp=self.min_disp,
                                                  max_disp=self.max_disp,
                                                  granularity=granularity,
                                                  random_epicenters=self.random_epicenters,
                                                  device=src.device)
            for k, v in six.iteritems(bundle):
                if 'field' in k:
                    #print ("Warp does nothing to {}".format(k))
                    pass
                elif 'tgt' in k:
                    #print (f"Warping {k} with {res_delta[0, 0].mean(), res_delta[0, 1].mean()}")
                    original_dtype = v.dtype
                    v = v.float()
                    v_aug = res_delta.from_pixels()(v)
                    if original_dtype == torch.bool:
                        v_aug = v_aug > 0.4
                    else:
                        zero_mask = v == 0
                        zero_mask_aug = res_delta.from_pixels()(zero_mask.float()) > 0.4
                        v_aug[zero_mask_aug] = 0
                    bundle[k] = v_aug
        return bundle

def generate_random_residuals(shape, max_disp, min_disp=0, granularity=9,
                              random_epicenters=True, device='cuda'):
    if random_epicenters:
        seed_shape = [i // (2**(granularity - 1)) for i in shape]
        up_shape = [i * 2 for i in shape]
    else:
        seed_shape = [i // (2**(granularity)) for i in shape]
        up_shape = shape
    seed_x = np.random.uniform(size=seed_shape, low=min_disp, high=max_disp)
    seed_y = np.random.uniform(size=seed_shape, low=min_disp, high=max_disp)

    up_x = skimage.transform.resize(seed_x, up_shape)
    up_y = skimage.transform.resize(seed_y, up_shape)

    final_x, final_y = random_crop([up_x, up_y], shape)
    result = torch.tensor(np.stack([final_x, final_y], axis=0), device=device)
    result = result.unsqueeze(0).float().field()
    return result

def get_random_crop_coords(full_shape, cropped_shape, coord_granularity=4):
    assert cropped_shape[0] <= full_shape[0]
    assert cropped_shape[1] <= full_shape[1]
    assert cropped_shape[0] % coord_granularity == 0
    assert cropped_shape[1] % coord_granularity == 0

    x_bot_preshift = np.random.randint(0, full_shape[0] - cropped_shape[0] + 1)
    y_bot_preshift = np.random.randint(0, full_shape[1] - cropped_shape[1] + 1)
    x_bot = x_bot_preshift - (x_bot_preshift % coord_granularity)
    y_bot = y_bot_preshift - (y_bot_preshift % coord_granularity)
    x_top = x_bot + cropped_shape[0]
    y_top = y_bot + cropped_shape[1]

    return x_bot, x_top, y_bot, y_top

def get_center_crop_coords(full_shape, cropped_shape, coord_granularity=4):
    assert cropped_shape[0] <= full_shape[0]
    assert cropped_shape[1] <= full_shape[1]

    assert cropped_shape[0] % coord_granularity == 0
    assert cropped_shape[1] % coord_granularity == 0

    x_bot_preshift = (full_shape[0] - cropped_shape[0]) // 2
    y_bot_preshift = (full_shape[1] - cropped_shape[1]) // 2
    x_bot = x_bot_preshift - (x_bot_preshift % coord_granularity)
    y_bot = y_bot_preshift - (y_bot_preshift % coord_granularity)

    x_top = x_bot + cropped_shape[0]
    y_top = y_bot + cropped_shape[1]

    return x_bot, x_top, y_bot, y_top

def random_crop(img, cropped_shape):
    result = []
    if isinstance(img, list):
        original_shape = img[0].shape[-2:]
        x_bot, x_top, y_bot, y_top = get_random_crop_coords(original_shape, cropped_shape)
        for i in img:
            assert (i.shape[-2] == original_shape[-2])
            assert (i.shape[-1] == original_shape[-1])

            result.append(i[..., x_bot:x_top, y_bot:y_top])
    else:
        original_shape = img.shape
        x_bot, x_top, y_bot, y_top = get_random_crop_coords(original_shape, cropped_shape)

        result.append(img[..., x_bot:x_top, y_bot:y_top])
    return result

class RandomTranspose(object):
    def __init__(self, prob=0.5):
        self.prob = prob

    def __call__(self, bundle):
        coin = np.random.uniform()
        if coin < self.prob:
            for k, v in six.iteritems(bundle):
                if 'field' not in k:
                    bundle[k] = v.transpose(-1, -2)
                else:
                    bundle['src_field'] = bundle['src_field'].transpose(2,3).flip(2)

        return bundle

class RandomSrcTgtSwap(object):
    def __init__(self, prob=0.5):
        self.prob = prob

    def __call__(self, bundle):
        result = copy.deepcopy(bundle)
        coin = np.random.uniform()
        if coin < self.prob:
            for k, v in six.iteritems(bundle):
                if 'src' in k:
                    result[k.replace('src', 'tgt')] = bundle[k]
                if 'tgt' in k:
                    result[k.replace('tgt', 'src')] = bundle[k]

        return result








import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

import os
import pathlib
import copy

import artificery
import torchfields

from metroem.finetuner import optimize_pre_post_ups

def finetune_field(src, tgt, pred_res_start, src_defects=None, tgt_defects=None, lr=18e-1, sm=300e0, num_iter=60):

    mse_keys_to_apply = {
        'src': [
            {
                'name': 'src_defects',
                'binarization': {'strat': 'eq', 'value': 0},
                'coarsen_ranges': [(1, 0)]
             }
            ],
        'tgt':[
            {
                'name': 'tgt_defects',
                'binarization': {'strat': 'eq', 'value': 0},
                'coarsen_ranges': [(1, 0)]
            }
        ]
    }

    sm_keys_to_apply = {
       "src": [
         {
             "name": "src_defects",
             "mask_value": 1.0e-5,
             "binarization": {"strat": "eq", "value": 0},
             'coarsen_ranges': [(1, 0)]
         }
       ],
       "tgt": [
            {
                'name': 'tgt_defects',
                'binarization': {'strat': 'eq', 'value': 0},
                'coarsen_ranges': [(1, 0)]
             }
       ]

   }


    src_small_defects = None
    src_large_defects = None

    if src_defects is not None:
        src_defects = src_defects.squeeze(0)
    else:
        src_defects = torch.zeros_like(src)

    if src_small_defects is not None:
        src_small_defects = src_small_defects.squeeze(0)
    else:
        src_small_defects = torch.zeros_like(src)

    if src_large_defects is not None:
        src_large_defects = src_large_defects.squeeze(0)
    else:
        src_large_defects = torch.zeros_like(src)

    if tgt_defects is not None:
        tgt_defects = tgt_defects.squeeze(0)
    else:
        tgt_defects = torch.zeros_like(src_defects)
    pred_res_opt = optimize_pre_post_ups(src, tgt, pred_res_start,
                                    src_defects=src_defects,
                                    tgt_defects=tgt_defects,
                                    crop=16, num_iter=num_iter,
                                    sm_keys_to_apply=sm_keys_to_apply,
                                    mse_keys_to_apply=mse_keys_to_apply,
                                    sm=sm, lr=lr)
    return pred_res_opt

def create_model(checkpoint_folder, device='cpu', checkpoint_name="checkpoint"):
    a = artificery.Artificery()

    spec_path = os.path.join(checkpoint_folder, "model_spec.json")
    my_p = a.parse(spec_path)

    checkpoint_path = os.path.join(checkpoint_folder,
            f"{checkpoint_name}.state.pth.tar")

    if not os.path.isfile(checkpoint_path):
        print ("creating new checkpiont...")
        return my_p

    load_my_state_dict(my_p,
            torch.load(checkpoint_path,
                map_location=torch.device(device)))
    my_p.name = checkpoint_name

    return my_p


def load_my_state_dict(model, state_dict, delta=0):
    own_state = model.state_dict()

    reinit_downmodules = []
    reinit_upmodules = []

    for layer_name, param in state_dict.items():
        layer_name_deltad = copy.copy(layer_name)
        if "level_downmodules" in layer_name:
            prefix = "level_downmodules."
            delta = 0
        elif "level_upmodules" in layer_name:
            prefix = "level_upmodules."
            delta = 0
        else:
            prefix = None

        if prefix is not None:
            level = int(layer_name[len(prefix)][:1])
            level_d = level + delta
            layer_name_deltad = prefix + str(level_d) + layer_name[len(prefix) + 1:]

        load_weights = True
        if layer_name_deltad not in own_state:
            load_weights = False

        if load_weights:
            if isinstance(param, Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            own_state[layer_name_deltad].copy_(param)


class Aligner(nn.Module):
    def __init__(self, model_folder, pass_field=True, checkpoint_name="checkpoint",
            finetune=False, finetune_iter=100, finetune_lr=1e-1, finetune_sm=30e0,
            train=False):
        super().__init__()

        this_folder = pathlib.Path(__file__).parent.absolute()
        self.net = create_model(model_folder, checkpoint_name=checkpoint_name)
        self.net.name = checkpoint_name
        self.finetune = finetune
        self.pass_field = pass_field
        self.finetune_iter = finetune_iter
        self.finetune_lr = finetune_lr
        self.finetune_sm = finetune_sm
        self.train = train

    def forward(self, src_img, tgt_img, src_agg_field=None, tgt_agg_field=None,
            src_folds=None, tgt_folds=None,
            finetune=None,
            finetune_iter=None,
            finetune_lr=None,
            finetune_sm=None,
            train=None,
            return_state=False,
            final_stage=False,
            **kwargs):

        if 'cuda' in str(src_img.device):
            self.net = self.net.cuda(src_img.device)
        else:
            self.net = self.net.cpu()

        while len(src_img.shape) < 4:
            src_img = src_img.unsqueeze(0)
            tgt_img = tgt_img.unsqueeze(0)

        if src_agg_field is not None:
            while len(src_agg_field.shape) < 4:
                src_agg_field = src_agg_field.unsqueeze(0)

            src_agg_field = src_agg_field.field().from_pixels()
            warped_src_img = src_agg_field(src_img)
            src_agg_field = src_agg_field.pixels()
        else:
            warped_src_img = src_img

        if self.pass_field:
            net_input = torch.cat((src_img, tgt_img), 1).float()
        else:
            net_input = torch.cat((warped_src_img, tgt_img), 1).float()

        if (train is None and self.train) or train == True:
            pred_res = self.net.forward(x=net_input, in_field=src_agg_field)
        else:
            with torch.no_grad():
                pred_res = self.net.forward(x=net_input, in_field=src_agg_field)
                #print (pred_res.abs().mean())

        if not self.pass_field and src_agg_field is not None:
            pred_res = pred_res.field().from_pixels()(src_agg_field).pixels()

        if finetune or (finetune is None and self.finetune):
            if finetune_iter is None:
                finetune_iter = self.finetune_iter
            if finetune_lr is None:
                finetune_lr = self.finetune_lr
            if finetune_sm is None:
                finetune_sm = self.finetune_sm
                if final_stage:
                    finetune_sm *= 10.0e0
            embeddings = self.net.state['up']['0']['output']
            src_opt = embeddings[0, 0:embeddings.shape[1]//2].unsqueeze(0).detach()
            tgt_opt = embeddings[0, embeddings.shape[1]//2:].unsqueeze(0).detach()

            src_defects = src_img == 0
            tgt_defects = tgt_img == 0
            #tgt_defects = None

            pred_res = finetune_field(src_opt, tgt_opt,
                    pred_res,
                    src_defects=src_defects,
                    tgt_defects=tgt_defects,
                    lr=finetune_lr,
                    num_iter=finetune_iter,
                    sm=finetune_sm)
        if return_state:
            return pred_res.field(), self.net.state
        else:
            return pred_res.field()

    def get_embeddings(self, img, level=0, preserve_zeros=False):
        img = img.squeeze()
        assert len(img.shape) == 2
        while len(img.shape) < 4:
            img = img.unsqueeze(0)

        net_input = torch.cat((img, img), 1).float()

        with torch.no_grad():
            self.net.forward(x=net_input)

        emb = self.net.state['up'][str(level)]['output']
        img_emb = emb[0, 1:emb.shape[1]//2]
        if preserve_zeros:
            mask =  (img == 0).float()
            while mask.shape[-1] > img_emb.shape[-1]:
                mask = torch.nn.functional.max_pool2d(mask, 2)
            mask = mask != 0
            img_emb[..., mask.squeeze()] = 0
        return img_emb

    def save_checkpoint(self, checkpoint_folder):
        path = os.path.join(checkpoint_folder, f"{self.net.name}.state.pth.tar")
        torch.save(self.net.state_dict(), path)

    def load_checkpoint(self, checkpoint_folder, map_location):
        """Load from checkpoint

        Args:
            checkpoint_folder (str): path to checkpoint
            map_lcation (dict): {'cuda:0' : 'cudaN'}
        """
        checkpoint_path = os.path.join(checkpoint_folder,
                                       f"{self.net.name}.state.pth.tar")
        if  os.path.isfile(checkpoint_path):
            load_my_state_dict(self,
                               torch.load(checkpoint_path,
                                          map_location=map_location))




import torch
import torchfields

import scipy
import numpy as np
import skimage
import h5py
import time
from skimage import feature
from metroem.helpers import get_np

from scipy.ndimage import convolve
from scipy.ndimage.measurements import label

from pdb import set_trace as st

def get_warped_mask_set(bundle, res, keys_to_apply, inv=False):
    prewarp_result = torch.ones((1, bundle['src'].shape[-2], bundle['src'].shape[-1]),
                        device=bundle['src'].device)
    res = res.squeeze()
    for settings in keys_to_apply:
        name = settings['name']
        if name in bundle:
            mask = bundle[name].squeeze()
            if 'fm' in settings and len(mask.shape) > 2:
                mask = mask[settings['fm']:settings['fm']+1]

            while len(mask.shape) < len(prewarp_result.shape):
                mask = mask.unsqueeze(0)

            if 'binarization' in settings:
                mask = binarize(mask, settings['binarization'])

            if 'mask_value' in settings:
                prewarp_result[mask != 1.0] = settings['mask_value']
            else:
                prewarp_result[mask != 1.0] = 0.0

            around_the_mask = mask
            if 'coarsen_ranges' in settings:
                for length, weight in settings['coarsen_ranges']:
                    if length > 10:
                        around_the_mask = coarsen_mask(around_the_mask, length)
                    else:
                        around_the_mask = coarsen_mask(around_the_mask, length)
                    prewarp_result[(around_the_mask != 1.0) * (prewarp_result == 1)] = weight

    if (res != 0).sum() > 0 and not inv:
        if res.shape[1] == 2:
            result = res(prewarp_result.float())
        else:
            result = res.from_pixels()(prewarp_result.float())
    else:
        result = prewarp_result
    return result


def binarize(img, bin_setting):
    if bin_setting['strat']== 'eq':
        result = (img == bin_setting['value'])
    elif bin_setting['strat']== 'neq':
        result = (img != bin_setting['value'])
    elif bin_setting['strat']== 'lt':
        result = (img < bin_setting['value'])
    elif bin_setting['strat'] == 'gt':
        result = (img > bin_setting['value'])
    elif bin_setting['strat']== 'between':
        result = ((img > bin_setting['range'][0]) * (img < bin_setting['range'][0]))
    elif bin_setting['strat']== 'not_between':
        result = ((img < bin_setting['range'][0]) + (img > bin_setting['range'][0])) > 0
    return result.float()


def get_mse_and_smoothness_masks(bundle, mse_keys_to_apply,
                                  sm_keys_to_apply, **kwargs):
    if 'src' not in mse_keys_to_apply:
        mse_keys_to_apply['src'] = []
    if 'tgt' not in mse_keys_to_apply:
        mse_keys_to_apply['tgt'] = []
    if 'src' not in sm_keys_to_apply:
        sm_keys_to_apply['src'] = []
    if 'tgt' not in sm_keys_to_apply:
        sm_keys_to_apply['tgt'] = []
    return get_warped_srctgt_mask(bundle, mse_keys_to_apply, sm_keys_to_apply)


def get_warped_srctgt_mask(bundle, mse_keys_to_apply, sm_keys_to_apply):
    src_shape = bundle['src'].shape
    if len(src_shape) == 4:
        mask_shape = (src_shape[0], 1, src_shape[2], src_shape[3])
        num_channels = src_shape[1]
        channel_dim = 1
    elif len(src_shape) == 3:
        mask_shape = (1, src_shape[1], src_shape[2])
        num_channels = src_shape[0]
        channel_dim = 0
    else:
        raise Exception(f"Unsupported image shape: {src_shape}")

    mask_mse = torch.ones(mask_shape, device=bundle['src'].device)
    mask_sm  = torch.ones(mask_shape, device=bundle['src'].device)
    pred_res = bundle['pred_res']

    if 'src' in mse_keys_to_apply:
        src_mask_mse = get_warped_mask_set(bundle, pred_res,
                                       mse_keys_to_apply['src'])
        mask_mse *= src_mask_mse

    if 'tgt' in mse_keys_to_apply:
        tgt_mask_mse = get_warped_mask_set(bundle, torch.zeros_like(pred_res),
                                       mse_keys_to_apply['tgt'])
        mask_mse *= tgt_mask_mse

    if 'src' in sm_keys_to_apply:
        src_mask_sm = get_warped_mask_set(bundle, pred_res,
                                       sm_keys_to_apply['src'])
        mask_sm *= src_mask_sm

    if 'tgt' in sm_keys_to_apply:
        tgt_mask_sm = get_warped_mask_set(bundle, torch.zeros_like(pred_res),
                                       sm_keys_to_apply['tgt'])
        mask_sm *= tgt_mask_sm

    if 'src_tgt_comb' in sm_keys_to_apply:
        src_comb_mask_sm = get_warped_mask_set(bundle, torch.zeros_like(pred_res),
                                       sm_keys_to_apply['src_tgt_comb']['src'])
        tgt_comb_mask_sm = get_warped_mask_set(bundle, torch.zeros_like(pred_res),
                                       sm_keys_to_apply['src_tgt_comb']['tgt'])
        mask_sm *= ((tgt_comb_mask_sm + src_comb_mask_sm) > 0).float()

    mask_mse = torch.cat([mask_mse] * num_channels, channel_dim)

    return mask_mse, mask_sm


RAW_WHITE_THRESHOLD = -0.485
def get_raw_defect_mask(img, threshold=-1):
    result = 1 - ((img < threshold) * (img > RAW_WHITE_THRESHOLD))
    return result.type(torch.cuda.FloatTensor)

def get_raw_white_mask(img):
    result = img >= RAW_WHITE_THRESHOLD
    return result.type(torch.cuda.FloatTensor)

def get_defect_mask(img, threshold=-3.5):
    result = 1 - ((img < threshold) * (img > -3.9999))
    return result.type(torch.cuda.FloatTensor)

def get_brightness_mask(img, low_cutoff, high_cutoff):
    result = (img >= low_cutoff)* (img <= high_cutoff)
    return result.type(torch.cuda.FloatTensor)

def get_blood_vessel_mask(img, threshold=2.5):
    result = img >= threshold
    return result.type(torch.cuda.FloatTensor)

def get_white_mask(img, threshold=-3.5):
    result = img >= threshold
    return result.type(torch.cuda.FloatTensor)

def get_black_mask(img):
    result = img < 3.55
    return result.type(torch.cuda.FloatTensor)


# Numpy masks
def get_very_white_mask(img):
    # expects each pixel in range [-0.5, 0.5]
    # used at mip 8
    return img > 0.04

def coarsen_mask(mask, n=1, flip=True):
    kernel = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
    for _ in range(n):
        if isinstance(mask, np.ndarray):
            mask = convolve(mask, kernel) > 0
            mask = mask.astype(np.int16) > 1
        else:
            kernel_var = torch.cuda.FloatTensor(kernel).unsqueeze(0).unsqueeze(0).to(mask.device)
            k = torch.nn.Parameter(data=kernel_var, requires_grad=False)
            if flip:
                mask = mask.logical_not().float()
            while len(mask.shape) < 4:
                mask = mask.unsqueeze(0)
            mask =  (torch.nn.functional.conv2d(mask,
                kernel_var, padding=1) > 1).squeeze(1)
            if flip:
                mask = mask.logical_not()
            mask = mask.float()
    return mask


def dilate(a):
    return scipy.ndimage.morphology.binary_dilation(a != 0)
    conn = scipy.ndimage.generate_binary_structure(2, 2)
    return scipy.ndimage.morphology.binary_dilation(a, conn)


def erode(a):
    return scipy.ndimage.morphology.binary_erosion(a != 0)


def closing(a, n=2):
    result = a
    for _ in range(n):
        result = dilate(result != 0)
    for _ in range(n):
        result = erode(result != 0)
    return result



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="metroem",
    version="0.0.5",
    author="Seung Lab",
    author_email="",
    description="Metric learning optimization pyramid for EM alignment",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    package_data={'': ['*.py']},
    install_requires=[
        'torch',
        'torchvision',
        'pathlib',
        'artificery',
        'torchfields',
        'itertools',
        'h5py',
        'argparse',
        'cloud-volume',
        'tqdm',
        'scikit-image',
        'six',
        'numpy'
    ],
    entry_points={
        "console_scripts": [
            "metroem-train = metroem.train:main"
        ],
    },
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
)



from pdb import set_trace as st
import glob
import h5py
import os
import re
import torch
import torchfields
import numpy as np

from collections import defaultdict

from metroem.alignment import align_sample
from metroem import helpers

class MyConcatDataset(torch.utils.data.ConcatDataset):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)
        self.size_limit = super().__len__()

    def __len__(self):
        return self.size_limit

    def set_size_limit(self, size_limit):
        self.size_limit = min(size_limit, super().__len__())

class SimpleH5Dataset(torch.utils.data.Dataset):
    def __init__(self, dset):
        self.dset = dset

    def __len__(self):
        return dset.shape[0]

    def __getitem__(self, i):
        return self.dset[i]


class MultimipDataset:
    """Access image & field samples by MIP for dataset directory

    Filename conventions within dataset directory:
      Image: '<NAME>_MIP>MIP.H5'
      Field: 'field_<STAGE>_<NAME>_<MIP>MIP_<CHECKPOINT NAME>.H5'
        <NAME>: x<X_START>_y<Y_START>_z<Z_START>

    Fields are associated with a checkpoint name, because different experiments
    may produce different fields as intermediaries between modules.
    """
    def __init__(self, path, aug_params=None, field_tag=None):
        self.path = path
        self.aug_params = aug_params
        self.data_piece_registry = {}
        self.min_mip = 9
        self.max_mip = 0
        self.names = set()

        if field_tag is None:
            self.field_tag = ''
        else:
            self.field_tag = "_" + field_tag

        self.img_dsets = defaultdict(lambda: defaultdict(lambda: None))
        self.field_dsets = defaultdict(lambda : defaultdict(lambda: defaultdict(lambda: None)))
        self.img_composite_dsets = dict()
        self.field_composite_dsets = defaultdict(lambda: defaultdict(lambda: None))
        self.load()


    def load(self):
        pattern = r"x(\d+)_y(\d+)_z(\d+)_(.*)MIP(\d).h5"
        files = glob.glob(os.path.join(self.path, "*.h5"))

        for f_abs in files:
            f = os.path.basename(f_abs)
            match = re.search(pattern, f)
            if match is not None:
                name = "x{}_y{}_z{}{}".format(match.group(1), match.group(2),
                                              match.group(3), match.group(4))
                mip = int(match.group(5))
                self.max_mip = max(mip, self.max_mip)
                self.min_mip = min(mip, self.min_mip)
                if name not in self.names:
                    self.names.add(name)
                    print (f"Adding '{name}' dataset.")

        for n in self.names:

            for mip in range(self.min_mip, self.max_mip + 1):
                self.load_img_dset(n, mip)
            '''
            files = glob.glob(os.path.join(self.path, f"field_*{n}*.h5"))
            pattern = re.compile(f"field_(\d+)_{name}_MIP(\d){self.field_tag}.h5")

            for f_abs in files:
                f = os.path.basename(f_abs)
                match = pattern.match(f)
                if match is not None:
                    stage = int(match.group(1))
                    mip = int(match.group(2))
                    self.load_field_dset(n, mip, stage)'''

    def load_img_composite_dset(self, mip):
        dset_list = []
        for name in self.names:
            dset_list.append(self.get_img_dset(mip=mip, name=name))
        if len(dset_list) == 0:
            raise Exceptoin(f"No image datasets found at MIP{mip}")

        self.img_composite_dsets[mip] = MyConcatDataset(dset_list)

    def load_field_composite_dset(self, mip, stage):
        dset_list = []
        for name in self.names:
            dset = self.get_field_dset(mip=mip, name=name, stage=stage)
            if dset is None:
                raise Exception(f"No field datasets with tag '{self.field_tag}' found for stage {stage} at MIP{mip} for dataset '{name}'")
            dset_list.append(dset)
        self.field_composite_dsets[mip][stage] = MyConcatDataset(dset_list)

    def get_img_dset(self, name, mip):
        return self.img_dsets[mip][name]

    def load_img_dset(self, name, mip):
        img_file = self.get_img_file(name=name, mip=mip)
        if img_file is None:
            return None

        if "img" in img_file:
            img_dset = img_file["img"]
        else:
            img_dset = img_file["main"]
        self.img_dsets[mip][name] = img_dset

    def get_field_dset(self, name, mip, stage):
        if self.field_dsets[mip][name][stage] is None:
            self.load_field_dset(name=name, mip=mip, stage=stage)
        return self.field_dsets[mip][name][stage]

    def load_field_dset(self, name, mip, stage, shape=None, create=False):
        """Return field datset for name, mip, stage; if not present, create it

        Args:
            name (str)
            mip (int)
            stage (int)
            shape (Tuple(int)): shape of associated image dataset, Nx2xWxH
            create (bool)
        """
        field_file = self.get_field_file(name=name, mip=mip, stage=stage, create=create)
        if field_file is None and not create:
            for prev_mip in range(mip+1, self.max_mip + 1):
                prev_field_file = self.get_field_file(name=name, mip=prev_mip, stage=stage)
                if prev_field_file is not None:
                    # self._upsample_field(name=name, stage=stage, mip_start=prev_mip, mip_end=mip)
                    self._upsample_field_direct(name=name,
                                                stage=stage,
                                                in_mip=prev_mip,
                                                out_mip=mip)
                    field_file = self.get_field_file(name=name, mip=mip, stage=stage, create=create)
                    break

        if field_file is None:
            return None

        field_dset = None
        if create:
            if "main" in field_file:
                del field_file["main"]
            if "field" in field_file:
                del field_file["field"]

            field_shape = (shape[0], 2, shape[-2], shape[-1])
            chunks = (1, 2, shape[-2], shape[-1])
            field_dset = field_file.create_dataset("main",
                    shape=field_shape,
                    dtype=np.float32,
                    chunks=chunks,
                    compression='lzf',
                    scaleoffset=2
                    )
        else:
            if "main" in field_file:
                field_dset = field_file["main"]
            elif "field" in field_file:
                field_dset = field_file["field"]

        self.field_dsets[mip][name][stage] = field_dset
        return field_dset

    def get_img_file(self, name, mip):
        path = os.path.join(self.path, f"{name}_MIP{mip}.h5")
        print (f"Loading file '{path}...'")
        if os.path.isfile(path):
            return h5py.File(path, 'r')
        else:
            return None

    def get_field_file(self, name, mip, stage, create=False):
        path = os.path.join(self.path, f"field_{stage}_{name}_MIP{mip}{self.field_tag}.h5")
        print (f"Loading file '{path}...'")
        if create:
            return h5py.File(path, 'a')
        if  os.path.isfile(path):
            return h5py.File(path, 'r')
        else:
            return None

    def generate_fields(self, model, mip, stage):
        for name in self.names:
            img_dset = self.get_img_dset(name, mip)
            if stage > 0:
                prev_field_file = self.get_field_file(name, mip, stage - 1)
                if prev_field_file is None:
                    raise Exception("Attempting to generate field for "\
                            f"stage {stage} before field for stage {stage-1} "\
                            f"has been generated (at MIP{mip})")
                if "field" in prev_field_file:
                    prev_field_dset = prev_field_file["field"]
                else:
                    prev_field_dset = prev_field_file["main"]

            else:
                prev_field_dset = None

            self.load_field_dset(name, mip, stage, shape=img_dset.shape, create=True)
            field_dset = self.get_field_dset(name, mip, stage)

            self._generate_field_dataset(model, img_dset, field_dset, prev_field_dset)

    def _generate_field_dataset(self, model, img_dset, field_dset, prev_field_dset):
        """Generate predicted source field for src, tgt pairs
        Pairs are permutable, so predict field for src,tgt as well as tgt,src
        """
        for b in range(img_dset.shape[0]):
            src = helpers.to_tensor(img_dset[b, 0])
            tgt = helpers.to_tensor(img_dset[b, 1])
            if prev_field_dset is not None:
                prev_field = helpers.to_tensor(prev_field_dset[b])
            else:
                prev_field = None

            field = model(src_img=src, tgt_img=tgt, src_agg_field=prev_field, train=False,
                    return_state=False)
            field_dset[b] = helpers.get_np(field)

    def _upsample_field(self, name, stage, mip_start, mip_end):
        for src_mip in range(mip_start, mip_end - 1, -1):
            tgt_mip = src_mip - 1

            src_field_dset = self.get_field_dset(name=name, stage=stage, mip=src_mip)
            tgt_img_dset = self.get_img_dset(name=name, mip=tgt_mip)
            tgt_field_dset = self.load_field_dset(name=name, stage=stage, mip=tgt_mip,
                                                  shape=tgt_img_dset.shape, create=True)

            with torch.no_grad():
                tgt_size = tgt_img_dset.shape[-1]

                for b in range(src_field_dset.shape[0]):
                    field_data = helpers.to_tensor(src_field_dset[b:b+1])
                    field_data_ups = torch.nn.functional.interpolate(field_data,
                                                                     mode='bilinear',
                                                                     scale_factor=2.0,
                                                                     align_corners=False,
                                                                     recompute_scale_factor=False
                                                                     ) * 2.0
                    field_data_ups_cropped = field_data_ups[:, :, :tgt_size, :tgt_size]
                    tgt_field_dset[b] = helpers.get_np(field_data_ups_cropped[...])

    def _upsample_field_direct(self, name, stage, in_mip, out_mip):
        """Upsample a field datset from SRC_MIP to DST_MIP

        Args:
            name (str): filename of dataset (input & output)
            stage (int): module which created dataset (input & output)
            in_mip (int): MIP level of the input dataset
            out_mip (int): MIP level of the output dataset
        """
        assert(in_mip > out_mip)
        in_field_dset = self.get_field_dset(name=name, stage=stage, mip=in_mip)
        dst_img_dset = self.get_img_dset(name=name, mip=out_mip)
        out_field_dset = self.load_field_dset(name=name,
                                              stage=stage,
                                              mip=out_mip,
                                              shape=dst_img_dset.shape,
                                              create=True)
        # scale_factor = 2**(in_mip - out_mip)

        with torch.no_grad():
            dst_size = dst_img_dset.shape[-1]
            for n in range(in_field_dset.shape[0]):
                in_field = helpers.to_tensor(in_field_dset[n:n+1]).field()
                in_field = in_field * (2**in_mip)
                out_field = in_field.up(mips=in_mip - out_mip)
                out_field = out_field / (2**out_mip)
                # out_field = torch.nn.functional.interpolate(in_field,
                #                                  mode='bilinear',
                #                                  scale_factor=scale_factor,
                #                                  align_corners=False,
                #                                  recompute_scale_factor=False
                #                                  ) * scale_factor
                out_field_cropped = out_field[0, :, :dst_size, :dst_size]
                out_field_dset[n] = helpers.get_np(out_field_cropped[...])


    def get_alignment_dset(self, mip, stage=None, start_index=0, end_index=None,
            crop_mode=None, cropped_size=None):
        """Create AlignmentDataLoader providing image pairs with associated fields

        Details about associated fields:
            * If STAGE > 0, then there must be a field available
            * The field available must be at the required MIP of STAGE
            * The field for the current STAGE is indexed at the current MIP,
                but by the previous STAGE (STAGE - 1)
        """

        if self.img_composite_dsets[mip] is None:
            self.load_img_composite_dset(mip)
        if stage not in [None, 0] and self.field_composite_dsets[mip][stage - 1] is None:
            self.load_field_composite_dset(mip, stage - 1)

        return AlignmentDataLoader(
                img_dset=self.img_composite_dsets[mip],
                field_dset=self.field_composite_dsets[mip][stage - 1],
                start_index=start_index,
                end_index=end_index,
                crop_mode=crop_mode,
                cropped_size=cropped_size
                )

    def get_train_dset(self, mip, stage=None, crop_mode='random', cropped_size=1024):
        """Get training dataset: all but last two image pairs
        """
        self.load_img_composite_dset(mip)
        return self.get_alignment_dset(mip=mip,
                               stage=stage,
                               start_index=0,
                               end_index=len(self.img_composite_dsets[mip])-2,
                               crop_mode=crop_mode,
                               cropped_size=cropped_size)

    def get_val_dset(self, mip, stage=None, crop_mode='middle', cropped_size=2048):
        """Get validation dataset: last two images pairs
        """
        self.load_img_composite_dset(mip)
        return self.get_alignment_dset(mip=mip,
                               stage=stage,
                               start_index=len(self.img_composite_dsets[mip])-2,
                               crop_mode=crop_mode,
                               cropped_size=cropped_size)


class AlignmentDataLoader(torch.utils.data.Dataset):
    def __init__(self,
                img_dset,
                field_dset,
                start_index,
                end_index,
                crop_mode=None,
                cropped_size=None):
        self.img_dset = img_dset
        self.field_dset = field_dset

        self.end_index = len(img_dset)
        self.start_index = 0
        if end_index is not None:
            self.end_index = min(self.end_index, end_index)
        if start_index is not None:
            self.start_index = max(self.start_index, start_index)
        self.shape = self.img_dset[0].shape
        self.crop_mode = crop_mode
        self.cropped_size = cropped_size

    def __len__(self):
        return self.end_index - self.start_index

    def set_size_limit(self, n):
        if n < len(self):
            self.end_index = self.start_index + n

    def _get_crop_coords(self):
        if self.crop_mode == 'random':
            x_bot, x_top, y_bot, y_top = get_random_crop_coords(self.shape[-2:], self.cropped_size)
        elif self.crop_mode == 'middle':
            x_bot, x_top, y_bot, y_top = get_center_crop_coords(self.shape[-2:], self.cropped_size)
        elif self.crop_mode is None:
            x_bot = y_bot = 0
            x_top = self.shape[-2]
            y_top = self.shape[-1]
        else:
            raise Exception(f"Bad crop mode: '{self.crop_mode}'")

        return x_bot, x_top, y_bot, y_top

    def __getitem__(self, i):
        x_bot, x_top, y_bot, y_top = self._get_crop_coords()
        x_slice = slice(x_bot, x_top)
        y_slice = slice(y_bot, y_top)

        img = self.img_dset[self.start_index + i]
        src = helpers.to_tensor(img[0, x_slice, y_slice])
        tgt = helpers.to_tensor(img[1, x_slice, y_slice])
        src[src < -4] = 0
        tgt[tgt < -4] = 0

        field = None
        if self.field_dset is not None:
            full_field = self.field_dset[self.start_index + i]
            field = helpers.to_tensor(full_field[...,x_bot:x_top, y_bot:y_top])

        bundle = {
            "src": src,
            "tgt": tgt,
            "src_zeros": src == 0,
            "tgt_zeros": tgt == 0,
                }
        if field is not None:
            bundle["src_field"] = field
        return bundle

def get_random_crop_coords(full_shape, cropped_size):
    if full_shape[0] < cropped_size:
        x_bot = 0
        x_top = full_shape[0]
    else:
        x_bot = np.random.randint(0, full_shape[0] - cropped_size + 1)
        x_top = x_bot + cropped_size

    if full_shape[1] < cropped_size:
        y_bot = 0
        y_top = full_shape[1]
    else:
        y_bot = np.random.randint(0, full_shape[1] - cropped_size + 1)
        y_top = y_bot + cropped_size

    return x_bot, x_top, y_bot, y_top

def get_center_crop_coords(full_shape, cropped_size):
    if full_shape[0] < cropped_size:
        x_bot = 0
        x_top = full_shape[0]
    else:
        x_bot = (full_shape[0] - cropped_size) // 2
        x_top = x_bot + cropped_size

    if full_shape[1] < cropped_size:
        y_bot = 0
        y_top = full_shape[1]
    else:
        y_bot = (full_shape[1] - cropped_size) // 2
        y_top = y_bot + cropped_size

    return x_bot, x_top, y_bot, y_top


